
/*

From an ontology defined as a set of Sentence objects, computes
a model that includes inferences and definitions of constructed
concepts.

Written by Richard Power
Copyright (c) 2013 by Open University

This package is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Library General Public License for more details.

*/

import java.lang.String;
import java.util.*;

public class Model
{ private ArrayList<ModelTerm> classList;
  private ArrayList<ModelTerm> propertyList;
  private ArrayList<Definition> defList;
  private ArrayList<Statement> statements;
  private ArrayList<Explanation> explanations;
  private ArrayList<Equivalence> equivalences;
  private ArrayList<Proof> proofs;
  private Explanation currentExp;
  private int distance;
  private int source;
  private Statement newStatement;

// Features
  public static final int CLASS = 1;
  public static final int INDIVIDUAL = 2;
  public static final int PROPERTY = 3;
  public static final int ORIGINAL = 4;
  public static final int TRANSITIVE = 5;
  public static final int RESTRICTION = 6;
  public static final int CONJUNCTION = 7;
  public static final int LACONIC = 8;
  public static final int DEFINED = 9;
  public static final int NOUN_PHRASE = 10;
  public static final int VERB_PHRASE = 11;
  public static final int VERB = 12;
  public static final int UNSPECIFIED = 13;
  public static final int INFERRED = 14;

  public Model(ArrayList<Sentence> sentences)
  { classList = new ArrayList<ModelTerm>();
    propertyList = new ArrayList<ModelTerm>();
    defList = new ArrayList<Definition>();
    statements = new ArrayList<Statement>();
    explanations = new ArrayList<Explanation>();
    equivalences = new ArrayList<Equivalence>();
    proofs = new ArrayList<Proof>();
    OwlTerm oterm;
    ModelTerm dummy = new ModelTerm(CLASS,"undef",ORIGINAL,NOUN_PHRASE);
    Statement statement = new Statement(dummy,dummy,ORIGINAL,0);
    distance = 0;
    source = ORIGINAL;
    for (Sentence sent : sentences)
    { if (sent.sentenceIsCorrect() == true)
      { oterm = sent.getOwlTerm();
        if ((oterm.getFunctor()==OwlTerm.SUB_CLASS_OF) ||
            (oterm.getFunctor()==OwlTerm.CLASS_ASSERTION))
        { ArrayList<OwlTerm> oterms = unpack(oterm);
          for (OwlTerm ot : oterms)
          { statement = refactor(ot);
            addStatement(statement);
          }
        }
      }
    }
    newStatement = statement;
    boolean modelComplete = false;
    while (modelComplete == false) modelComplete = checkModel();
    checkEquivalence();
    constructProofs();
    findLaconicStatements();
  }

// Make a model for a list of sentences which incorporates any classes found
// in a new sentence, but without adding the new statement itself. This is
// needed for model diffing.
 
  public Model(ArrayList<Sentence> sentences, Sentence newsent)
  { classList = new ArrayList<ModelTerm>();
    propertyList = new ArrayList<ModelTerm>();
    defList = new ArrayList<Definition>();
    statements = new ArrayList<Statement>();
    explanations = new ArrayList<Explanation>();
    equivalences = new ArrayList<Equivalence>();
    proofs = new ArrayList<Proof>();
    OwlTerm oterm;
    ModelTerm dummy = new ModelTerm(CLASS,"undef",ORIGINAL,NOUN_PHRASE);
    Statement statement = new Statement(dummy,dummy,ORIGINAL,0);
    distance = 0;
    source = ORIGINAL;
    for (Sentence sent : sentences)
    { if (sent.sentenceIsCorrect())
      { oterm = sent.getOwlTerm();
        if ((oterm.getFunctor()==OwlTerm.SUB_CLASS_OF) ||
            (oterm.getFunctor()==OwlTerm.CLASS_ASSERTION))
        { ArrayList<OwlTerm> oterms = unpack(oterm);
          for (OwlTerm ot : oterms)
          { statement = refactor(ot);
            addStatement(statement);
          }
        }
      }
    }
    if (newsent.sentenceIsCorrect())
    { oterm = newsent.getOwlTerm();
      if ((oterm.getFunctor()==OwlTerm.SUB_CLASS_OF) ||
          (oterm.getFunctor()==OwlTerm.CLASS_ASSERTION))
      statement = refactor(oterm);
    }
    newStatement = statement;
    boolean modelComplete = false;
    while (modelComplete == false) modelComplete = checkModel();
    checkEquivalence();
    constructProofs();
    findLaconicStatements();
  }

  public ArrayList<ModelTerm> getClasses() { return classList; }
  public ArrayList<ModelTerm> getProperties() { return propertyList; }
  public ArrayList<Definition> getDefinitions() { return defList; }
  public ArrayList<Statement> getStatements() { return statements; }
  public ArrayList<Explanation> getExplanations() { return explanations; }
  public ArrayList<Equivalence> getEquivalences() { return equivalences; }
  public Statement getNewStatement() { return newStatement; }

  private Statement refactor(OwlTerm oterm)
  { ModelTerm arg0 = refactorClass(oterm.getArgument(0));
    ModelTerm arg1 = refactorClass(oterm.getArgument(1));
    Statement stat = new Statement(arg0,arg1,ORIGINAL,0);
    return stat;
  }

  private ModelTerm refactorClass(OwlTerm oterm)
  { int type = CLASS;
    int source = ORIGINAL;
    int category = NOUN_PHRASE;
    ModelTerm mterm = new ModelTerm(CLASS,"undef",ORIGINAL,NOUN_PHRASE);
    ModelTerm arg0, arg1;
    String name = "undef";
    String propname = "undef";
    if (oterm.getFunctor() == OwlTerm.NAMED_INDIVIDUAL)
    { type = INDIVIDUAL;
      name = checkCapital(oterm.getEntityName());
      if (questionTerm(oterm)==false) mterm = addClass(type,name,source,category);
    }
    else if (oterm.getFunctor() == OwlTerm.CLASS)
    { name = oterm.getEntityName();
      mterm = addClass(type,name,source,category);
    }
    else if (oterm.getFunctor() == OwlTerm.OBJECT_SOME_VALUES_FROM)
    { propname = oterm.getArgument(0).getEntityName();
      arg0 = addProperty(propname);
      arg1 = refactorClass(oterm.getArgument(1));
      name = propname + " " + indefiniteArticle(arg1.getName()) + " " + arg1.getName();
      source = DEFINED;
      category = VERB_PHRASE;
      mterm = addClass(type,name,source,category);
      addDefinition(RESTRICTION,mterm,arg0,arg1);
    }
    else if (oterm.getFunctor() == OwlTerm.OBJECT_HAS_VALUE)
    { propname = oterm.getArgument(0).getEntityName();
      arg0 = addProperty(propname);
      arg1 = refactorClass(oterm.getArgument(1));
      name = propname + " " + arg1.getName();
      source = DEFINED;
      category = VERB_PHRASE;
      mterm = addClass(type,name,source,category);
      addDefinition(RESTRICTION,mterm,arg0,arg1);
    }

//  Conjunctions of the form "thing that VP" simplify to "VP"
    else if ((oterm.getFunctor() == OwlTerm.OBJECT_INTERSECTION_OF) &&
             (oterm.getArgument(0).getEntityName().equals("thing")))
    mterm = refactorClass(oterm.getArgument(1));

    else if (oterm.getFunctor() == OwlTerm.OBJECT_INTERSECTION_OF)
    { arg0 = refactorClass(oterm.getArgument(0));
      arg1 = refactorClass(oterm.getArgument(1));
      name = arg0.getName();
      if (arg0.getCategory() == NOUN_PHRASE)
      name = name + " that ";
      else name = name + " and ";
      if (arg1.getCategory() == NOUN_PHRASE)
      name = name + "is " + indefiniteArticle(arg1.getName()) + " " + arg1.getName();
      else name = name + arg1.getName();
      source = DEFINED;
      category = arg0.getCategory();
      mterm = addClass(type,name,source,category);
      addDefinition(CONJUNCTION,mterm,arg0,arg1);
    }
    return mterm;
  }

  private boolean questionTerm(OwlTerm oterm)
  { boolean result = false;
    String name = oterm.getEntityName();
    if (name.equals("What") || name.equals("Who")) result = true;
    return result;
  }

  private ModelTerm addClass(int type, String name, int source, int category)
  { ModelTerm mterm = new ModelTerm(type,name,source,category);
    boolean found = false;
    for (ModelTerm c : classList)
    if ((c.getType()==type) &&
        (c.getName().equals(name)) &&
        (c.getSource()==source) &&
        (c.getCategory()==category))
    { found = true; mterm = c; }
    if (found == false) classList.add(mterm);
    return mterm;
  }

  private ModelTerm addProperty(String name)
  { ModelTerm mterm = new ModelTerm(PROPERTY,name,ORIGINAL,VERB);
    boolean found = false;
    for (ModelTerm p : propertyList)
    if (p.getName().equals(name)) 
    { found = true; mterm = p; }
    if (found == false) propertyList.add(mterm);
    return mterm;
  }

  private void addDefinition(int type, ModelTerm mterm, ModelTerm arg0, ModelTerm arg1)
  { Definition newdef = new Definition(mterm,type,arg0,arg1);
    boolean found = false;
    for (Definition def : defList)
    if ( (def.getType()==type) &&
         (def.getDefinedTerm().equals(mterm)))
    found = true;
    if (found == false) defList.add(newdef);
  }

  private void addStatement(Statement statement)
  { boolean found = false;
    for (Statement s : statements)
    if ( (s.getSubject()==statement.getSubject()) &&
         (s.getPredicate()==statement.getPredicate()))
    found = true;
    if (found == false) statements.add(statement);
  }

// Changes "The ..." to "the ..."
  private String checkCapital(String name)
  { if (name.startsWith("The "))
    name = name.replaceFirst("The","the");
    return name;
  }
  
// Try to add a new entailment, returning true when model is complete

  private boolean checkModel()
  { for (ModelTerm c1 : classList)
    for (ModelTerm c2 : classList)
    if (newEntailment(c1,c2))
    { Statement stat = new Statement(c1,c2,source,distance);
      currentExp.setEntailment(stat);
      statements.add(stat);
      explanations.add(currentExp);
      return false;
    }
    return true;
  }

  private boolean newEntailment(ModelTerm class1, ModelTerm class2)
  { distance = 0;
    if (class1.equals(class2)) return false;
    else
    { Statement stat = getStatement(class1,class2);
      if (existingStatement(stat)) return false;
      else if (laconicProof(class1,class2)) return true;
      else if (conjunctionProof(class1,class2)) return true;
      else if (restrictionProof(class1,class2)) return true;
      else if (transitiveProof(class1,class2)) return true;
      else return false;
    }
  }

// If no matching statement is found, return the UNSPECIFIED one
  private Statement getStatement(ModelTerm class1, ModelTerm class2)
  { Statement stat = new Statement(class1,class2,UNSPECIFIED,0);
    for (Statement s : statements)
    if ((s.getSubject().equals(class1)) &&
        (s.getPredicate().equals(class2)))
    stat = s;
    return stat;
  }

  private boolean existingStatement(Statement stat)
  { if (stat.getSource()==UNSPECIFIED) return false;
    else return true;
  }

  private boolean transitiveProof(ModelTerm class1, ModelTerm class2)
  { for (Statement stat1 : statements)
    if (stat1.getSubject().equals(class1))
    { Statement stat2 = getStatement(stat1.getPredicate(),class2);
      if (existingStatement(stat2))
      { currentExp = new Explanation(TRANSITIVE);
        currentExp.setStatement1(stat1);
        currentExp.setStatement2(stat2);
        distance = stat1.getDistance() + stat2.getDistance() + 1;
        source = TRANSITIVE;
        return true;
      }
    }
    return false;
  }
  
  private boolean laconicProof(ModelTerm class1, ModelTerm class2)
  { for (Definition def : defList)
    if (def.getDefinedTerm().equals(class1) && (def.getType()==CONJUNCTION))
    if (def.getArg0().equals(class2) || def.getArg1().equals(class2))
    { currentExp = new Explanation(LACONIC);
      currentExp.setDefinition1(def);
      distance = 1;
      source = LACONIC;
      return true;
    }
    return false;
  }
  
  private boolean conjunctionProof(ModelTerm class1, ModelTerm class2)
  { for (Definition def : defList)
    if (class2.getSource()==DEFINED)
    if ((def.getType()==CONJUNCTION) && def.getDefinedTerm().equals(class2))
    { Statement stat1 = getStatement(class1,def.getArg0());
      if (existingStatement(stat1))
      { Statement stat2 = getStatement(class1,def.getArg1());
        if (existingStatement(stat2))
        { currentExp = new Explanation(CONJUNCTION);
          currentExp.setStatement1(stat1);
          currentExp.setStatement2(stat2);
          currentExp.setDefinition1(def);
          distance = stat1.getDistance() + stat2.getDistance() + 1;
          source = CONJUNCTION;
          return true;
        }
      }
    }
    return false;
  }
  
  private boolean restrictionProof(ModelTerm class1, ModelTerm class2)
  { if ((class1.getSource()==DEFINED) && (class2.getSource()==DEFINED))
    { Definition def1 = getDefinition(class1);
      Definition def2 = getDefinition(class2);
      if ((def1.getType()==RESTRICTION) && (def2.getType()==RESTRICTION))
      if (def1.getArg0().equals(def2.getArg0()))
      { Statement stat = getStatement(def1.getArg1(),def2.getArg1());
        if (existingStatement(stat))
        { currentExp = new Explanation(RESTRICTION);
          currentExp.setDefinition1(def1);
          currentExp.setDefinition2(def2);
          currentExp.setStatement1(stat);
          distance = stat.getDistance() + 1;
          source = RESTRICTION;
          return true;
        }
      }
    }
    return false;
  }

// If there is no matching definition the UNSPECIFIED one is returned
  private Definition getDefinition(ModelTerm mterm)
  { Definition def = new Definition(mterm,UNSPECIFIED,mterm,mterm);
    for (Definition d : defList)
    if (d.getDefinedTerm().equals(mterm)) def = d;
    return def;
  }

// For each statement check whether its inverse also holds
  private void checkEquivalence()
  { int size = statements.size();
    for (int i=0; i<size; i++)
    { Statement stat = statements.get(i);
      ModelTerm subj = stat.getSubject();
      ModelTerm pred = stat.getPredicate();
      for (int j=i+1; j<size; j++)
      { Statement inv = statements.get(j);
        if (inv.getSubject().equals(pred) && inv.getPredicate().equals(subj))
        { Equivalence equiv = new Equivalence(subj,pred);
          equivalences.add(equiv);
        }
      }
    }
  }

// For each explanation build a complete proof
  private void constructProofs()
  { for (Statement stat : statements)
    { Proof proof = new Proof(stat,explanations);
      proofs.add(proof);
//      System.out.println(proof.report());
    }
  }

// Replace OwlTerm with set of terms, unpacking top-level intersections

  private ArrayList<OwlTerm> unpack(OwlTerm oterm)
  { ArrayList<OwlTerm> oterms = new ArrayList<OwlTerm>();
    OwlTerm subj = oterm.getArgument(0);
    OwlTerm pred = oterm.getArgument(1);
    int functor = oterm.getFunctor();
    OwlTerm newterm;
    while (pred.getFunctor()==OwlTerm.OBJECT_INTERSECTION_OF)
    { newterm = new OwlTerm(functor,2);
      newterm.setArgument(0,subj);
      newterm.setArgument(1,pred.getArgument(0));
      oterms.add(newterm);
      pred = pred.getArgument(1);
    }
    newterm = new OwlTerm(functor,2);
    newterm.setArgument(0,subj);
    newterm.setArgument(1,pred);
    oterms.add(newterm);
    return oterms;
  }

  private void findLaconicStatements()
  { for (Statement stat : statements)
    { boolean laconic = true;
      ModelTerm pred = stat.getPredicate();
      for (Definition def : defList)
      if (def.getDefinedTerm().equals(pred) && (def.getType()==CONJUNCTION))
      laconic = false;
      stat.setLaconic(laconic);
    }
  }

  private String indefiniteArticle(String noun)
  { String indef = "a";
    if ((noun.charAt(0)=='a') ||
        (noun.charAt(0)=='e') ||
        (noun.charAt(0)=='i') ||
        (noun.charAt(0)=='o') ||
        (noun.charAt(0)=='u') ||
        (noun.charAt(0)=='A') ||
        (noun.charAt(0)=='E') ||
        (noun.charAt(0)=='I') ||
        (noun.charAt(0)=='O') ||
        (noun.charAt(0)=='U'))
    indef = "an";
    return indef;
  }
  
  public String report()
  { String s = "";
    if (! classList.isEmpty()) s = s + "INDIVIDUALS AND CLASSES\n";
    for (ModelTerm c : classList) s = s + c.getReportName() + "\n";
    if (! propertyList.isEmpty()) s = s + "PROPERTIES\n";
    for (ModelTerm p : propertyList) s = s + p.getName() + "\n";
    if (! defList.isEmpty()) s = s + "DEFINITIONS\n";
    for (Definition d : defList) s = s + d.report() + "\n";
    if (! statements.isEmpty()) s = s + "AXIOMS\n";
    for (Statement stat : statements)
    if (stat.getSource()==ORIGINAL) s = s + stat.reportNL() + "\n";
    if (! statements.isEmpty()) s = s + "ENTAILMENTS\n";
    for (Statement stat : statements)
    if (stat.getSource()!=ORIGINAL) s = s + stat.reportNL() + "\n";
    if (! explanations.isEmpty()) s = s + "EXPLANATIONS\n";
    for (Explanation exp : explanations) s = s + exp.report() + "\n";
    if (! equivalences.isEmpty()) s = s + "EQUIVALENCES\n";
    for (Equivalence equiv : equivalences) s = s + equiv.report() + "\n";
    return s;
  }

  public String reportEntailments()
  { String s = "";
    for (Statement stat : statements)
    if (stat.getSource()!=ORIGINAL) s = s + stat.reportNL() + ".\n";
    return s;
  }

  public String explanationResponse()
  { String s = "Okay. ";
    for (Explanation e : explanations)
    { Statement stat = e.getEntailment();
      if ((stat.getSubject().equals(newStatement.getSubject())) &&
          (stat.getPredicate().equals(newStatement.getPredicate())))
      s = e.reportDialogue();
    }
    return s;
  }

  public String proofResponse()
  { String s = "Okay. ";
    for (Proof p : proofs)
    { Statement stat = p.getNode();
      if ((stat.getSubject().equals(newStatement.getSubject())) &&
          (stat.getPredicate().equals(newStatement.getPredicate())))
      s = p.reportNL();
    }
    return s;
  }

  public String retrievalResponse()
  { String s = "";
    for (Statement stat : statements)
    if (stat.getPredicate().equals(newStatement.getPredicate()))
    { String phrase = stat.getSubject().getName();
      if (stat.getSubject().getType()==CLASS)
      { if (stat.getSubject().getCategory()==NOUN_PHRASE)
        phrase = indefiniteArticle(phrase) + " " + phrase;
        else phrase = "a thing that " + phrase;
      }
      s = s + phrase + ", ";
    }
    if (s.equals("")) s = "Nothing.";
    else s = "The following: " + s.substring(0,s.length()-2) + ".";
    return s;
  }

  public String descriptionResponse()
  { String s = "";
    String np = newStatement.getSubject().getName();
    if (newStatement.getSubject().getType()==CLASS)
    { if (newStatement.getSubject().getCategory()==NOUN_PHRASE)
      np = indefiniteArticle(np) + " " + np;
      else np = "A thing that " + np;
      if (np.startsWith("a")) np = np.replaceFirst("a","A");
    }
    if (np.startsWith("the")) np = np.replaceFirst("the","The");
    for (Statement stat : statements)
    if (stat.getSubject().equals(newStatement.getSubject()))
    { String phrase = stat.getPredicate().getName();
      if (stat.getPredicate().getType()==CLASS)
      { if (stat.getPredicate().getCategory()==NOUN_PHRASE)
        phrase = indefiniteArticle(phrase) + " " + phrase;
        else phrase = "a thing that " + phrase;
      }
      s = s + phrase + ", ";
    }
    if (s.equals("")) s = "No information.";
    else 
    s = np + " is all of the following: " + s.substring(0,s.length()-2) + ".";
    return s;
  }

  public void display()
  { System.out.print("Statement: ");
    System.out.println("List of classes");
    for (ModelTerm mterm : classList) { mterm.display(); System.out.println(); }
    System.out.println("List of properties");
    for (ModelTerm mterm : propertyList) { mterm.display(); System.out.println(); }
    System.out.println("List of definitions");
    for (Definition def : defList) def.display();
    System.out.println("List of statements");
    for (Statement stat : statements) stat.display();
  }

}
