
/*

Represents the difference between two models

Written by Richard Power
Copyright (c) 2013 by Open University

This package is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Library General Public License for more details.

*/

import java.lang.String;
import java.util.*;

public class ModelDiff
{ private ArrayList<ModelTerm> classesAdded, classesRemoved;
  private ArrayList<ModelTerm> propertiesAdded, propertiesRemoved;
  private ArrayList<Definition> definitionsAdded, definitionsRemoved;
  private ArrayList<Statement> statementsAdded, statementsRemoved;
  private ArrayList<Statement> axiomsAdded, axiomsRemoved;
  private ArrayList<Statement> entailmentsAdded, entailmentsRemoved;
  private ArrayList<Explanation> explanationsAdded, explanationsRemoved;
  private ArrayList<Equivalence> equivalencesAdded, equivalencesRemoved;
  private Model model1, model2;

  public ModelDiff(Model arg1, Model arg2)
  { model1 = arg1;
    model2 = arg2;

    classesAdded = diffModelTerms(model1.getClasses(),model2.getClasses());
    propertiesAdded = diffModelTerms(model1.getProperties(),model2.getProperties());
    definitionsAdded = diffDefinitions(model1.getDefinitions(),model2.getDefinitions());;
    statementsAdded = diffStatements(model1.getStatements(),model2.getStatements());
    explanationsAdded = diffExplanations(model1.getExplanations(),model2.getExplanations());
    equivalencesAdded = diffEquivalences(model1.getEquivalences(),model2.getEquivalences());

    classesRemoved = diffModelTerms(model2.getClasses(),model1.getClasses());
    propertiesRemoved = diffModelTerms(model2.getProperties(),model1.getProperties());
    definitionsRemoved = diffDefinitions(model2.getDefinitions(),model1.getDefinitions());;
    statementsRemoved = diffStatements(model2.getStatements(),model1.getStatements());
    explanationsRemoved = diffExplanations(model2.getExplanations(),model1.getExplanations());
    equivalencesRemoved = diffEquivalences(model2.getEquivalences(),model1.getEquivalences());

    axiomsAdded = new ArrayList<Statement>();
    axiomsRemoved = new ArrayList<Statement>();
    entailmentsAdded = new ArrayList<Statement>();
    entailmentsRemoved = new ArrayList<Statement>();

    for (Statement s : statementsAdded)
    if (s.getSource()==Model.ORIGINAL) axiomsAdded.add(s);
    else entailmentsAdded.add(s);

    for (Statement s : statementsRemoved)
    if (s.getSource()==Model.ORIGINAL) axiomsRemoved.add(s);
    else entailmentsRemoved.add(s);
  }

  private ArrayList<ModelTerm> diffModelTerms(ArrayList<ModelTerm> model1,
                                              ArrayList<ModelTerm> model2)
  { ArrayList<ModelTerm> diff = new ArrayList<ModelTerm>();
    for (ModelTerm mterm : model1)
    if (equivalentModelTerm(mterm,model2)==false)
    diff.add(mterm);
    return diff;
  }

// Assume equivalent if name and type are the same
  private boolean equivalentModelTerm(ModelTerm mterm, ArrayList<ModelTerm> mlist)
  { boolean equiv = false;
    for (ModelTerm m : mlist)
    if ((mterm.getType()==m.getType()) && mterm.getName().equals(m.getName()))
    equiv = true;
    return equiv;
  }

  private ArrayList<Definition> diffDefinitions(ArrayList<Definition> model1,
                                                ArrayList<Definition> model2)
  { ArrayList<Definition> diff = new ArrayList<Definition>();
    for (Definition def : model1)
    if (equivalentDefinition(def,model2)==false)
    diff.add(def);
    return diff;
  }

// Assume equivalent if name of defined term and type are the same
  private boolean equivalentDefinition(Definition def, ArrayList<Definition> deflist)
  { boolean equiv = false;
    for (Definition d : deflist)
    if ((def.getType()==d.getType()) && def.getDefinedTerm().getName().equals(d.getDefinedTerm().getName()))
    equiv = true;
    return equiv;
  }

  private ArrayList<Statement> diffStatements(ArrayList<Statement> model1,
                                              ArrayList<Statement> model2)
  { ArrayList<Statement> diff = new ArrayList<Statement>();
    for (Statement stat : model1)
    if (equivalentStatement(stat,model2)==false)
    diff.add(stat);
    return diff;
  }

// Assume equivalent if subjects and predicates are equivalent modelterms
  private boolean equivalentStatement(Statement stat, ArrayList<Statement> statlist)
  { boolean equiv = false;
    for (Statement s : statlist)
    if ((stat.getSubject().getType() == s.getSubject().getType()) &&
        stat.getSubject().getName().equals(s.getSubject().getName()) &&
        (stat.getPredicate().getType() == s.getPredicate().getType()) &&
        stat.getPredicate().getName().equals(s.getPredicate().getName()))
    equiv = true;
    return equiv;
  }

  private ArrayList<Explanation> diffExplanations(ArrayList<Explanation> model1,
                                                  ArrayList<Explanation> model2)
  { ArrayList<Explanation> diff = new ArrayList<Explanation>();
    for (Explanation exp : model1)
    if (equivalentExplanation(exp,model2)==false)
    diff.add(exp);
    return diff;
  }

// Assume equivalent if subjects and predicates of entailments are equivalent modelterms and rules are same
  private boolean equivalentExplanation(Explanation exp, ArrayList<Explanation> explist)
  { boolean equiv = false;
    Statement stat = exp.getEntailment();
    for (Explanation e : explist)
    { Statement s = e.getEntailment();
      if ((stat.getSubject().getType() == s.getSubject().getType()) &&
          stat.getSubject().getName().equals(s.getSubject().getName()) &&
          (stat.getPredicate().getType() == s.getPredicate().getType()) &&
          stat.getPredicate().getName().equals(s.getPredicate().getName()) &&
          (exp.getRule() == e.getRule()))
      equiv = true;
    }
    return equiv;
  }
  
  private ArrayList<Equivalence> diffEquivalences(ArrayList<Equivalence> model1,
                                                  ArrayList<Equivalence> model2)
  { ArrayList<Equivalence> diff = new ArrayList<Equivalence>();
    for (Equivalence equiv : model1)
    if (equivalentEquivalence(equiv,model2)==false)
    diff.add(equiv);
    return diff;
  }

// Assume equivalent if terms are equivalent modelterms
  private boolean equivalentEquivalence(Equivalence eq, ArrayList<Equivalence> eqlist)
  { boolean equiv = false;
    for (Equivalence e : eqlist)
    { if ((e.getTerm1().getType() == eq.getTerm1().getType()) &&
          e.getTerm1().getName().equals(eq.getTerm1().getName()) &&
          (e.getTerm2().getType() == eq.getTerm2().getType()) &&
          e.getTerm2().getName().equals(eq.getTerm2().getName()))
      equiv = true;
      if ((e.getTerm1().getType() == eq.getTerm2().getType()) &&
          e.getTerm1().getName().equals(eq.getTerm2().getName()) &&
          (e.getTerm2().getType() == eq.getTerm1().getType()) &&
          e.getTerm2().getName().equals(eq.getTerm1().getName()))
      equiv = true;
    }
    return equiv;
  }

  public String report()
  { String s = "";
    if (! classesAdded.isEmpty()) s = s + "CLASSES ADDED\n";
    for (ModelTerm c : classesAdded) s = s + c.getReportName() + "\n";
    if (! classesRemoved.isEmpty()) s = s + "CLASSES REMOVED\n";
    for (ModelTerm c : classesRemoved) s = s + c.getReportName() + "\n";
    if (! propertiesAdded.isEmpty()) s = s + "PROPERTIES ADDED\n";
    for (ModelTerm p : propertiesAdded) s = s + p.getName() + "\n";
    if (! propertiesRemoved.isEmpty()) s = s + "PROPERTIES REMOVED\n";
    for (ModelTerm p : propertiesRemoved) s = s + p.getName() + "\n";
    if (! definitionsAdded.isEmpty()) s = s + "DEFINITIONS ADDED\n";
    for (Definition d : definitionsAdded) s = s + d.report() + "\n";
    if (! definitionsRemoved.isEmpty()) s = s + "DEFINITIONS REMOVED\n";
    for (Definition d : definitionsRemoved) s = s + d.report() + "\n";
    if (! statementsAdded.isEmpty()) s = s + "STATEMENTS ADDED\n";
    for (Statement stat : statementsAdded) s = s + stat.report() + "\n";
    if (! statementsRemoved.isEmpty()) s = s + "STATEMENTS REMOVED\n";
    for (Statement stat : statementsRemoved) s = s + stat.report() + "\n";
    if (! explanationsAdded.isEmpty()) s = s + "EXPLANATIONS ADDED\n";
    for (Explanation exp : explanationsAdded) s = s + exp.report() + "\n";
    if (! explanationsRemoved.isEmpty()) s = s + "EXPLANATIONS REMOVED\n";
    for (Explanation exp : explanationsRemoved) s = s + exp.report() + "\n";
    if (! equivalencesAdded.isEmpty()) s = s + "EQUIVALENCES ADDED\n";
    for (Equivalence eq : equivalencesAdded) s = s + eq.report() + "\n";
    if (! equivalencesRemoved.isEmpty()) s = s + "EQUIVALENCES REMOVED\n";
    for (Equivalence eq : equivalencesRemoved) s = s + eq.report() + "\n";
    return s;
  }

  public String reportNL()
  { String s = "";
    String m = "\u2212 ";
    if ((! classesAdded.isEmpty()) || (! classesRemoved.isEmpty()))
    s = s + "INDIVIDUALS AND CLASSES\n";
    for (ModelTerm c : classesAdded) s = s + "+ " + c.getReportName() + "\n";
    for (ModelTerm c : classesRemoved) s = s + m + c.getReportName() + "\n";
    if ((! propertiesAdded.isEmpty()) || (! propertiesRemoved.isEmpty()))
    s = s + "PROPERTIES\n";
    for (ModelTerm p : propertiesAdded) s = s + "+ " + p.getName() + "\n";
    for (ModelTerm p : propertiesRemoved) s = s + m + p.getName() + "\n";
    if ((! axiomsAdded.isEmpty())|| (! axiomsRemoved.isEmpty()))
    s = s + "AXIOMS\n";
    for (Statement stat : axiomsAdded) s = s + "+ " + stat.reportNL() + "\n";
    for (Statement stat : axiomsRemoved) s = s + m + stat.reportNL() + "\n";
    if ((! entailmentsAdded.isEmpty())|| (! entailmentsRemoved.isEmpty()))
    s = s + "ENTAILMENTS\n";
    for (Statement stat : entailmentsAdded) s = s + "+ " + stat.reportNL() + "\n";
    for (Statement stat : entailmentsRemoved) s = s + m + stat.reportNL() + "\n";
    if ((! equivalencesAdded.isEmpty())|| (! equivalencesRemoved.isEmpty()))
    s = s + "EQUIVALENCES\n";
    for (Equivalence eq : equivalencesAdded) s = s + "+ " + eq.report() + "\n";
    for (Equivalence eq : equivalencesRemoved) s = s + m + eq.report() + "\n";
    if (! explanationsAdded.isEmpty())
    s = s + "EXPLANATIONS\n";
    for (Explanation e : explanationsAdded) s = s + e.report() + "\n";
    return s;
  }

// Response with salient entailment
  public String response()
  { String s = "";
    if (statementsAdded.isEmpty()) s = "Already inferred.";
    else if (entailmentsAdded.isEmpty()) s = "Okay.";
    else if (! equivalencesAdded.isEmpty())
    { TreeSet<String> names = getClassNames(equivalencesAdded);
      s = "So the following are equivalent: ";
      for (String name : names) s = s + name + ", ";
      s = s.substring(0,s.length()-2) + ".";
    }
    else
    { Statement best = entailmentsAdded.get(0);
      ModelTerm pred = statementsAdded.get(0).getPredicate();
      for (Statement stat : entailmentsAdded)
      if (stat.getSource() < best.getSource())
      best = stat;
      else if ((stat.getSource()==best.getSource()) &&
               (predScore(stat,pred)>predScore(best,pred)))
      best = stat;
      else if ((stat.getSource()==best.getSource()) &&
               (predScore(stat,pred)==predScore(best,pred)) &&
               (stat.getDistance() > best.getDistance()))
      best = stat;
      if (best.getSource()!=Model.LACONIC) s = "So " + best.reportPhrase();
      else s = "Okay.";
    }
    return s;
  }

// Response with all entailments listed
  public String responseFull()
  { String s = "";
    if (statementsAdded.isEmpty()) s = "Already inferred.";
    else if (! equivalencesAdded.isEmpty())
    { TreeSet<String> names = getClassNames(equivalencesAdded);
      s = "So the following are equivalent: ";
      for (String name : names) s = s + name + ", ";
      s = s.substring(0,s.length()-2) + ".";
    }
    else if (entailmentsAdded.isEmpty()) s = "Okay.";
    else
    { s = "Okay, so:";
      for (Statement stat : entailmentsAdded)
      if (stat.isLaconic()) s = s + "\n - " + stat.reportNL();
    }
    return s;
  }

  public String retractionResponse()
  { String s = "";
    if (entailmentsRemoved.isEmpty()) s = "Okay.";
    else if (! equivalencesRemoved.isEmpty())
    { TreeSet<String> names = getClassNames(equivalencesRemoved);
      s = "So it is unknown whether the following are equivalent: ";
      for (String name : names) s = s + name + ", ";
      s = s.substring(0,s.length()-2) + ".";
    }
    else
    { Statement best = entailmentsRemoved.get(0);
      ModelTerm pred = statementsRemoved.get(0).getPredicate();
      for (Statement stat : entailmentsRemoved)
      if (stat.getSource() < best.getSource())
      best = stat;
      else if ((stat.getSource()==best.getSource()) &&
               (predScore(stat,pred)>predScore(best,pred)))
      best = stat;
      else if ((stat.getSource()==best.getSource()) &&
               (predScore(stat,pred)==predScore(best,pred)) &&
               (stat.getDistance() > best.getDistance()))
      best = stat;
      if (best.getSource()!=Model.LACONIC)
      s = "So it is unknown whether " + best.reportPhrase();
      else s = "Okay.";
    }
    return s;
  }

  public String retractionResponseFull()
  { String s = "";
    if (! equivalencesRemoved.isEmpty())
    { TreeSet<String> names = getClassNames(equivalencesRemoved);
      s = "So it is unknown whether the following are equivalent: ";
      for (String name : names) s = s + name + ", ";
      s = s.substring(0,s.length()-2) + ".";
    }
    else if (entailmentsRemoved.isEmpty()) s = "Okay.";
    else
    { s = "Okay, so the following no longer hold:";
      for (Statement stat : entailmentsRemoved)
      if (stat.isLaconic()) s = s + "\n - " + stat.reportNL();
    }
    return s;
  }

  public TreeSet<String> getClassNames(ArrayList<Equivalence> eqlist)
  { TreeSet<String> names = new TreeSet<String>();
    for (Equivalence equiv : eqlist)
    { names.add(equiv.getTerm1().getName());
      names.add(equiv.getTerm2().getName());
    }
    return names;
  }

  private int predScore(Statement stat, ModelTerm mterm)
  { if (stat.getPredicate().equals(mterm))
    return 1;
    else return 0;
  }

}
