
/*

Implements the equivalent of a Prolog term with functor and arguments.
Functors are restricted to a subset of OWL functors (roughly EL) and
can thus be represented by integer constants rather than by strings.

Written by Richard Power
Copyright (c) 2013 by Open University

This package is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Library General Public License for more details.

*/

import java.util.Vector;
import java.util.Collections;
import java.lang.String;

public class OwlTerm
{ private int functor;
  private OwlTerm[] argument;
  private int numberOfArguments;
  private String entityName;
  private Vector<Word> word;
  private int headWord; 

// The head of a class or property name is the noun/verb that is
// inflected to distinguish singular from plural. It is set externally
// depending on the position of the word in the name. For class names
// it is the final noun or the noun preceding the first preposition. For
// property names it is the first word in the name.

// OWL functors
  public static final int UNSPECIFIED = 0;
  public static final int ANNOTATION_ASSERTION = 1;
  public static final int ANNOTATION_PROPERTY = 2;
  public static final int ANNOTATION_PROPERTY_DOMAIN = 3;
  public static final int ANNOTATION_PROPERTY_RANGE = 4;
  public static final int ANONYMOUS_INDIVIDUAL = 5;
  public static final int ASYMMETRIC_OBJECT_PROPERTY = 6;
  public static final int CLASS = 7;
  public static final int CLASS_ASSERTION = 8;
  public static final int DATA_ALL_VALUES_FROM = 9;
  public static final int DATA_COMPLEMENT_OF = 10;
  public static final int DATA_EXACT_CARDINALITY = 11;
  public static final int DATA_HAS_VALUE = 12;
  public static final int DATA_INTERSECTION_OF = 13;
  public static final int DATA_MAX_CARDINALITY = 14;
  public static final int DATA_MIN_CARDINALITY = 15;
  public static final int DATA_ONE_OF = 16;
  public static final int DATA_PROPERTY = 17;
  public static final int DATA_PROPERTY_ASSERTION = 18;
  public static final int DATA_PROPERTY_DOMAIN = 19;
  public static final int DATA_PROPERTY_RANGE = 20;
  public static final int DATA_SOME_VALUES_FROM = 21;
  public static final int DATA_UNION_OF = 22;
  public static final int DATATYPE = 23;
  public static final int DATATYPE_DEFINITION = 24;
  public static final int DATATYPE_RESTRICTION = 25;
  public static final int DIFFERENT_INDIVIDUALS = 26;
  public static final int DISJOINT_CLASSES = 27;
  public static final int DISJOINT_DATA_PROPERTIES = 28;
  public static final int DISJOINT_OBJECT_PROPERTIES = 29;
  public static final int DISJOINT_UNION = 30;
  public static final int EQUIVALENT_CLASSES = 31;
  public static final int EQUIVALENT_DATA_PROPERTIES = 32;
  public static final int EQUIVALENT_OBJECT_PROPERTIES = 33;
  public static final int FUNCTIONAL_DATA_PROPERTY = 34;
  public static final int FUNCTIONAL_OBJECT_PROPERTY = 35;
  public static final int HAS_KEY = 36;
  public static final int INVERSE_FUNCTIONAL_OBJECT_PROPERTY = 37;
  public static final int INVERSE_OBJECT_PROPERTIES = 38;
  public static final int IRREFLEXIVE_OBJECT_PROPERTY = 39;
  public static final int LITERAL = 40;
  public static final int NAMED_INDIVIDUAL = 41;
  public static final int NEGATIVE_DATA_PROPERTY_ASSERTION = 42;
  public static final int NEGATIVE_OBJECT_PROPERTY_ASSERTION = 43;
  public static final int OBJECT_ALL_VALUES_FROM = 44;
  public static final int OBJECT_COMPLEMENT_OF = 45;
  public static final int OBJECT_EXACT_CARDINALITY = 46;
  public static final int OBJECT_EXISTS_SELF = 47;
  public static final int OBJECT_HAS_SELF = 48;
  public static final int OBJECT_HAS_VALUE = 49;
  public static final int OBJECT_INTERSECTION_OF = 50;
  public static final int OBJECT_INVERSE_OF = 51;
  public static final int OBJECT_MAX_CARDINALITY = 52;
  public static final int OBJECT_MIN_CARDINALITY = 53;
  public static final int OBJECT_ONE_OF = 54;
  public static final int OBJECT_PROPERTY = 55;
  public static final int OBJECT_PROPERTY_ASSERTION = 56;
  public static final int OBJECT_PROPERTY_CHAIN = 57;
  public static final int OBJECT_PROPERTY_DOMAIN = 58;
  public static final int OBJECT_PROPERTY_RANGE = 59;
  public static final int OBJECT_SOME_VALUES_FROM = 60;
  public static final int OBJECT_UNION_OF = 61;
  public static final int REFLEXIVE_OBJECT_PROPERTY = 62;
  public static final int SAME_INDIVIDUAL = 63;
  public static final int SUB_ANNOTATION_PROPERTY_OF = 64;
  public static final int SUB_CLASS_OF = 65;
  public static final int SUB_DATA_PROPERTY_OF = 66;
  public static final int SUB_OBJECT_PROPERTY_OF = 67;
  public static final int SYMMETRIC_OBJECT_PROPERTY = 68;
  public static final int TRANSITIVE_OBJECT_PROPERTY = 69;
// Functors for comments message and metadata (outside OWL)
  public static final int COMMENT = 70;
  public static final int METADATA = 71;  

// Print names for each functor
  public static String functorString[] =
    { "Unspecified",
      "AnnotationAssertion",
      "AnnotationProperty",
      "AnnotationPropertyDomain",
      "AnnotationPropertyRange",
      "AnonymousIndividual",
      "AsymmetricObjectProperty",
      "Class",
      "ClassAssertion",
      "DataAllValuesFrom",
      "DataComplementOf",
      "DataExactCardinality",
      "DataHasValue",
      "DataIntersectionOf",
      "DataMaxCardinality",
      "DataMinCardinality",
      "DataOneOf",
      "DataProperty",
      "DataPropertyAssertion",
      "DataPropertyDomain",
      "DataPropertyRange",
      "DataSomeValuesFrom",
      "DataUnionOf",
      "Datatype",
      "DatatypeDefinition",
      "DatatypeRestriction",
      "DifferentIndividuals",
      "DisjointClasses",
      "DisjointDataProperties",
      "DisjointObjectProperties",
      "DisjointUnion",
      "EquivalentClasses",
      "EquivalentDataProperties",
      "EquivalentObjectProperties",
      "FunctionalDataProperty",
      "FunctionalObjectProperty",
      "HasKey",
      "InverseFunctionalObjectProperty",
      "InverseObjectProperties",
      "IrreflexiveObjectProperty",
      "Literal",
      "NamedIndividual",
      "NegativeDataPropertyAssertion",
      "NegativeObjectPropertyAssertion",
      "ObjectAllValuesFrom",
      "ObjectComplementOf",
      "ObjectExactCardinality",
      "ObjectExistsSelf",
      "ObjectHasSelf",
      "ObjectHasValue",
      "ObjectIntersectionOf",
      "ObjectInverseOf",
      "ObjectMaxCardinality",
      "ObjectMinCardinality",
      "ObjectOneOf",
      "ObjectProperty",
      "ObjectPropertyAssertion",
      "ObjectPropertyChain",
      "ObjectPropertyDomain",
      "ObjectPropertyRange",
      "ObjectSomeValuesFrom",
      "ObjectUnionOf",
      "ReflexiveObjectProperty",
      "SameIndividual",
      "SubAnnotationPropertyOf",
      "SubClassOf",
      "SubDataPropertyOf",
      "SubObjectPropertyOf",
      "SymmetricObjectProperty",
      "TransitiveObjectProperty",
      "Comment",
      "Metadata"
    };

// Main method used only for testing
  public static void main(String[] args)
  { OwlTerm term = new OwlTerm(SUB_CLASS_OF,2);
    OwlTerm arg0 = new OwlTerm(CLASS,0);
    OwlTerm arg1 = new OwlTerm(CLASS,0);
    arg0.setEntityName("dog");
    arg1.setEntityName("animal");
    term.setArgument(0,arg0);
    term.setArgument(1,arg1);
    term.display();
    term.setFunctor(CLASS_ASSERTION);
    arg0.setFunctor(NAMED_INDIVIDUAL);
    arg0.setEntityName("Rover");
    arg1.setEntityName("corgi");
    term.display();
  }

  public OwlTerm(int arg1, int arg2)
  { functor = arg1;
    numberOfArguments = arg2;
    if (arg2 > 0) argument = new OwlTerm[arg2];
    else word = new Vector<Word>();
    entityName = "";
  }
  
  public void setFunctor(int arg1) { functor = arg1; }
  public void setEntityName(String arg1) { entityName = arg1; }
  public void setArgument(int arg1, OwlTerm arg2) { argument[arg1] = arg2; }

  public int getNumberOfArguments() { return numberOfArguments; }
  public int getFunctor() { return functor; }
  public String getEntityName() { return entityName; }
  public OwlTerm getArgument(int arg1) { return argument[arg1]; }

// Update headWord and entityName every time a new word arrives in the vector
  public void addWord(Word w)
  { word.addElement(w);
    updateHeadWord();
    updateEntityName();
  }

  public void updateEntityName()
  { entityName = word.elementAt(0).getWordString();
    for (int i=1; i<word.size(); i++)
    entityName = entityName + " " + word.elementAt(i).getWordString();
  }

  public void updateHeadWord()
  { if (functor == CLASS) updateClassHeadWord();
    else headWord = 0;
  }

  public void updateClassHeadWord()
  { headWord = 0;
    while ((headWord<word.size()-1) && 
           (word.elementAt(headWord+1).getWordCategory()!=Word.PREPOSITION))
    headWord++;
  }

  public String getPrintForm()
  { String s = functorString[functor] + "(";
    if (entityNameFunctor(functor)) s = "";
    if (!entityName.equals(""))
    { if (numberOfArguments > 0) s = s + digitise(entityName) + " ";
      else s = s + makeIRI(functor,entityName);
    }
    if (functor==CLASS_ASSERTION)    
    { if (argument[1] == null) s = s + "null ";
      else s = s + argument[1].getPrintForm() + " ";
      if (argument[0] == null) s = s + "null ";
      else s = s + argument[0].getPrintForm();
    }
    else if (numberOfArguments > 0)
    { for (int i=0; i<numberOfArguments; i++)
      { if (i > 0) s = s + " ";
        if (argument[i] == null) s = s + "null";
        else s = s + argument[i].getPrintForm();
      }
    }
   if (!entityNameFunctor(functor)) s = s + ")";
   return s;
  }

  private boolean entityNameFunctor(int fn)
  { boolean result = false;
    if ((fn==CLASS)||(fn==NAMED_INDIVIDUAL)||(fn==OBJECT_PROPERTY))
    result = true;
    return result;
  } 

  private String digitise(String integer)
  { String s = integer;
    for (int i=0; i<Word.integerWord.length; i++)
    if (Word.integerWord[i].equals(integer)) s = "" + i;
    return s;
  }

  private String makeIRI(int functor, String name)
  { String iri = name;
    if ((functor != LITERAL) &&
        (functor != COMMENT) &&
        (functor != METADATA))
    iri = ":" + name.replace(' ','_');
    if (iri.equals("#thing")) iri = "owl:Thing";
    else if (iri.equals("#integer-value")) iri = "http://www.w3.org/2001/XMLSchema#integer";
    else if (iri.equals("#decimal-value")) iri = "http://www.w3.org/2001/XMLSchema#decimal";
    else if (iri.equals("#string-value")) iri = "http://www.w3.org/2001/XMLSchema#string";
    return iri;
  }

  public void display()
  { String printForm = getPrintForm();
    System.out.println(printForm);
  }

}
