
/*

Holds a proof tree as a node (explanation) and list of daughters (proofs)

Written by Richard Power
Copyright (c) 2013 by Open University

This package is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Library General Public License for more details.

===============================================================

Rule is TRANSITIVE|LACONIC|CONJUNCTION|RESTRICTION

Different rules make use of different variables:

TRANSITIVE: two statements
LACONIC: one definition
CONJUNCTION: one definition and two statements
RESTRICTION: two definitions and one statement

*/

import java.lang.String;
import java.util.*;

public class Proof
{ private Statement node;
  private int numberOfConstituents;
  private int rule;
  private boolean terminal;
  private ArrayList<Proof> cset;
  private ArrayList<Explanation> explanations;
  private ArrayList<Statement> lemmas;

  public Proof(Statement arg, ArrayList<Explanation> list)
  { node = arg; 
    explanations = list;
    cset = new ArrayList<Proof>();
    terminal = false;
    rule = Model.UNSPECIFIED;
    makeConstituents();
    numberOfConstituents = cset.size();
    lemmas = new ArrayList<Statement>();
    makeLemmas();
  }

  public Statement getNode() {return node; }
  public int getNumberOfConstituents() {return numberOfConstituents; }
  public Proof getConstituent(int i) {return cset.get(i); }
  public ArrayList<Proof> getConstituents() {return cset; }
  public ArrayList<Statement> getLemmas() {return lemmas; }

  private void makeConstituents()
  { if (node.getSource()==Model.ORIGINAL) terminal = true;
    else
    { Explanation exp = findExplanation();
      rule = exp.getRule();
      Proof pf;
      if ((rule==Model.TRANSITIVE) || (rule==Model.CONJUNCTION))
      { cset.add(new Proof(exp.getStatement1(),explanations));
        cset.add(new Proof(exp.getStatement2(),explanations));
      }
      else if (rule == Model.RESTRICTION)
      { cset.add(new Proof(exp.getStatement1(),explanations));
      }
    }
  }

  private Explanation findExplanation()
  { Explanation exp = new Explanation(0);
    for (Explanation e : explanations)
    if (e.getEntailment().equals(node)) exp = e;
    return exp;
  }

  private void makeLemmas()
  { if (terminal==false)
    { lemmas.add(node);
      for (Proof p : cset)
      for (Statement stat : p.getLemmas())
      if (!alreadyListed(stat)) lemmas.add(stat);
    }
  }

  private boolean alreadyListed(Statement stat)
  { boolean listed = false;
    for (Statement lem : lemmas)
    if (lem.equals(stat)) listed = true;
    return listed;
  }

  public String ruleString()
  { String s = "";
    if (rule==Model.UNSPECIFIED) s = "AXIOM";
    if (rule==Model.TRANSITIVE) s = "TRANS";
    if (rule==Model.LACONIC) s = "LACON";
    if (rule==Model.CONJUNCTION) s = "CONJU";
    if (rule==Model.RESTRICTION) s = "RESTR";
    return s;
  }

  public String report()
  { String s = "";
    if (numberOfConstituents > 0)
    { s = s + node.reportNL() + " because:";
      for (Proof p : cset) s = s + "\n - " + p.getNode().reportNL();
      for (Proof p : cset) s = s + "\n" + p.report();
    }
    return s;
  }

  public String reportNL()
  { String s = "";
    for (Statement stat : lemmas)
    for (Explanation exp : explanations)
    if (exp.getEntailment().equals(stat))
    { s = s + stat.reportNL() + " because:";
      s = s + "\n - " + exp.getStatement1().reportNL();
      if (exp.getRule()!=Model.RESTRICTION)
      s = s + "\n - " + exp.getStatement2().reportNL();
      s = s + "\n";
    }
    if (s.equals("")) s = "Asserted.\n";
    return s;
  }

  private void display(int indent)
  { for (int i=1; i<=indent; i++) System.out.print(" ");
    System.out.println(node.reportNL() + " (" + ruleString() + ")");
    for (Proof p : cset) p.display(indent+2);
  }

  public void display()
  { display(0);
    System.out.println();
  }

}
