
/*
Instances of Sentence hold information about sentences in the edited text

Written by Richard Power
Copyright (c) 2013 by Open University

This package is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Library General Public License for more details.

======================================================================

When sentences are constructed they are given their number in the text,
(counting from zero), their string content (possibly including '\r' on
Windows, but excluding '\n'), and their starting position within the
whole text string.

Sentence objects are responsible for calculating their own end positions
within the whole text string, tokenising the sentence string into words,
and analysing the sentence. The analysis consists of features attached to
the word objects: each word has a category, such as noun or verb, and a
state, which is the node in the finite-state transducer which is reached
after consuming the word.

Sentence objects return messages based on the state of the last completed
word. These are obtained by calling a getMessage method on that word object.

Sentences are split into words by space separators. This requires some
preliminary work on the sentence string, which may contain the following
problems:
-- Line feed character '\r' at end under Windows
-- Leading or trailing spaces
-- Extra spaces between consecutive words
-- Punctuation (especially full stop) directly following words
Thus if we get the input " Rover   is a dog.\r" we need to revise the
string to "Rover is a dog ." before a split based on space will work.

*/

import java.util.Vector;
import java.util.Collections;

public class Sentence
{ private String textString;
  private int numberOfCharacters, numberOfWords;
  private String[] wordStrings;
  private Word[] word;
  private OwlTerm axiom;
  private boolean correctSentence;
  private int speechAct;

// States of the transducer
  public static final int REJECT = 0;
  public static final int START = 1;
  public static final int ISUB = 2;
  public static final int CSUB = 3;
  public static final int SUBJECT = 4;
  public static final int COPULAR = 5;
  public static final int AUX = 6;
  public static final int NPLIST = 7;
  public static final int COMPLEMENT = 8;
  public static final int NPLISTLOOP = 9;
  public static final int VPLIST = 10;
  public static final int ILIST = 11;
  public static final int ILISTOBJ = 12;
  public static final int CLIST = 13;
  public static final int CLISTOBJ = 14;
  public static final int VPLISTLOOP = 15;
  public static final int AUXLIST = 16;
  public static final int VPCHAINLOOP = 17;
  public static final int AUXCHAIN = 18;
  public static final int VPCHAIN = 19;
  public static final int ICHAIN = 20;
  public static final int ICHAINOBJ = 21;
  public static final int CCHAIN = 22;
  public static final int CCHAINOBJ = 23;
  public static final int FINISH = 24;
// Additional states for speech acts
  public static final int SPEECHACT = 25;
  public static final int UNSPECIFIED1 = 26;
  public static final int UNSPECIFIED2 = 27;
// Additional states for complex subjects
  public static final int CSUBJECT = 28;
  public static final int SUBVPLIST = 29;
  public static final int SUBILIST = 30;
  public static final int SUBILISTOBJ = 31;
  public static final int SUBCLIST = 32;
  public static final int SUBCLISTOBJ = 33;
  public static final int SUBVPLISTLOOP = 34;
  public static final int SUBAUXLIST = 35;
  public static final int SUBVPCHAINLOOP = 36;
  public static final int SUBAUXCHAIN = 37;
  public static final int SUBVPCHAIN = 38;
  public static final int SUBICHAIN = 39;
  public static final int SUBICHAINOBJ = 40;
  public static final int SUBCCHAIN = 41;
  public static final int SUBCCHAINOBJ = 42;
// Additional states for "that is a" construction
  public static final int SUBCOPLIST = 43;
  public static final int SUBCOPCHAIN = 44;
  public static final int COPLIST = 45;
  public static final int COPCHAIN = 46;

// Roles for syntax highlighting
  public static final int ERROR = 0;
  public static final int SUGAR = 1;
  public static final int INDIVIDUAL = 2;
  public static final int CLASS = 3;
  public static final int PROPERTY = 4;
  public static final int LITERAL = 5;
  public static final int RETAIN = 6;
  public static final int REVISE = 7;
  public static final int SUGARBACK = 8;

// Roles for speech acts
  public static final int ASSERT = 0;
  public static final int RETRACT = 1;
  public static final int CONFIRM = 2;
  public static final int EXPLAIN = 3;
  public static final int RETRIEVE = 4;
  public static final int DESCRIBE = 5;

// Each transition is assigned a role for colour-coding 
  private static int role[][] =
   {{ START, ISUB, INDIVIDUAL },
    { START, CSUB, SUGAR },
    { START, SUBJECT, SUGARBACK },
    { START, SPEECHACT, SUGAR },
    { START, START, SUGAR },
    { SPEECHACT, START, SUGAR },

    { ISUB, SUBJECT, INDIVIDUAL },
    { CSUB, CSUBJECT, CLASS },
    { SUBJECT, SUBJECT, RETAIN },
    { CSUBJECT, CSUBJECT, RETAIN },
    { SUBJECT, COPULAR, SUGAR },
    { SUBJECT, AUX, PROPERTY },
    { SUBJECT, VPLIST, PROPERTY },
    { SUBJECT, FINISH, PROPERTY },
    { CSUBJECT, COPULAR, SUGAR },
    { CSUBJECT, AUX, PROPERTY },
    { CSUBJECT, VPLIST, PROPERTY },
    { SUBJECT, FINISH, PROPERTY },
    { CSUBJECT, SUBVPLISTLOOP, SUGAR },

    { COPULAR, NPLIST, SUGAR },
    { COPULAR, VPLIST, REVISE },
    { NPLIST, COMPLEMENT, CLASS },
    { COMPLEMENT, COMPLEMENT, CLASS },
    { COMPLEMENT, VPLISTLOOP, SUGAR },
    { COMPLEMENT, FINISH, SUGAR },
    { AUX, VPLIST, PROPERTY },
    { VPLIST, VPLIST, PROPERTY },
    { VPLIST, ILIST, INDIVIDUAL },
    { VPLIST, CLIST, SUGAR },
    { VPLIST, ILISTOBJ, INDIVIDUAL },
    { ILIST, ILISTOBJ, INDIVIDUAL },
    { ILISTOBJ, ILISTOBJ, INDIVIDUAL },
    { CLIST, CLISTOBJ, CLASS },
    { CLISTOBJ, CLISTOBJ, CLASS },
    { ILISTOBJ, FINISH, SUGAR },
    { CLISTOBJ, FINISH, SUGAR },
    { ILISTOBJ, VPLISTLOOP, SUGAR },
    { CLISTOBJ, VPLISTLOOP, SUGAR },
    { VPLISTLOOP, AUXLIST, PROPERTY },
    { AUXLIST, VPLIST, PROPERTY },
    { VPLISTLOOP, VPLIST, PROPERTY },
    { VPLIST, VPLIST, PROPERTY },

    { VPLISTLOOP, COPLIST, SUGAR },
    { COPLIST, CLIST, SUGAR },
    { COPLIST, VPLIST, REVISE },

    { SUBVPLIST, SUBVPLIST, PROPERTY },
    { SUBVPLIST, SUBILIST, INDIVIDUAL },
    { SUBVPLIST, SUBCLIST, SUGAR },
    { SUBVPLIST, SUBILISTOBJ, INDIVIDUAL },
    { SUBILIST, SUBILISTOBJ, INDIVIDUAL },
    { SUBILISTOBJ, SUBILISTOBJ, INDIVIDUAL },
    { SUBCLIST, SUBCLISTOBJ, CLASS },
    { SUBCLISTOBJ, SUBCLISTOBJ, CLASS },
    { SUBILISTOBJ, COPULAR, SUGAR },
    { SUBILISTOBJ, AUX, PROPERTY },
    { SUBILISTOBJ, VPLIST, PROPERTY },
    { SUBILISTOBJ, FINISH, PROPERTY },
    { SUBCLISTOBJ, COPULAR, SUGAR },
    { SUBCLISTOBJ, AUX, PROPERTY },
    { SUBCLISTOBJ, VPLIST, PROPERTY },
    { SUBCLISTOBJ, FINISH, PROPERTY },
    { SUBILISTOBJ, SUBVPLISTLOOP, SUGAR },
    { SUBCLISTOBJ, SUBVPLISTLOOP, SUGAR },
    { SUBVPLISTLOOP, SUBAUXLIST, PROPERTY },
    { SUBAUXLIST, SUBVPLIST, PROPERTY },
    { SUBVPLISTLOOP, SUBVPLIST, PROPERTY },
    { SUBVPLIST, SUBVPLIST, PROPERTY },

    { SUBVPLISTLOOP, SUBCOPLIST, SUGAR },
    { SUBCOPLIST, SUBCLIST, SUGAR },
    { SUBCOPLIST, SUBVPLIST, REVISE },

    { CLISTOBJ, VPCHAINLOOP, SUGAR },
    { VPCHAINLOOP, AUXCHAIN, PROPERTY },
    { VPCHAINLOOP, VPCHAIN, PROPERTY },
    { AUXCHAIN, VPCHAIN, PROPERTY },
    { VPCHAIN, ICHAIN, INDIVIDUAL },
    { VPCHAIN, CCHAIN, SUGAR },
    { VPCHAIN, ICHAINOBJ, INDIVIDUAL },
    { VPCHAIN, VPCHAIN, PROPERTY },
    { ICHAIN, ICHAINOBJ, INDIVIDUAL },
    { ICHAINOBJ, ICHAINOBJ, INDIVIDUAL },
    { CCHAIN, CCHAINOBJ, CLASS },
    { CCHAINOBJ, CCHAINOBJ, CLASS },
    { ICHAINOBJ, FINISH, SUGAR },
    { CCHAINOBJ, FINISH, SUGAR },
    { CCHAINOBJ, VPCHAINLOOP, SUGAR },

    { VPCHAINLOOP, COPCHAIN, SUGAR },
    { COPCHAIN, CCHAIN, SUGAR },
    { COPCHAIN, VPCHAIN, REVISE },

    { SUBCLISTOBJ, SUBVPCHAINLOOP, SUGAR },
    { SUBVPCHAINLOOP, SUBAUXCHAIN, PROPERTY },
    { SUBVPCHAINLOOP, SUBVPCHAIN, PROPERTY },
    { SUBAUXCHAIN, SUBVPCHAIN, PROPERTY },
    { SUBVPCHAIN, SUBICHAIN, INDIVIDUAL },
    { SUBVPCHAIN, SUBCCHAIN, SUGAR },
    { SUBVPCHAIN, SUBICHAINOBJ, INDIVIDUAL },
    { SUBVPCHAIN, SUBVPCHAIN, PROPERTY },
    { SUBICHAIN, SUBICHAINOBJ, INDIVIDUAL },
    { SUBICHAINOBJ, SUBICHAINOBJ, INDIVIDUAL },
    { SUBCCHAIN, SUBCCHAINOBJ, CLASS },
    { SUBCCHAINOBJ, SUBCCHAINOBJ, CLASS },
    { SUBICHAINOBJ, COPULAR, SUGAR },
    { SUBICHAINOBJ, AUX, PROPERTY },
    { SUBICHAINOBJ, VPLIST, PROPERTY },
    { SUBICHAINOBJ, FINISH, SUGAR },
    { SUBCCHAINOBJ, COPULAR, SUGAR },
    { SUBCCHAINOBJ, AUX, PROPERTY },
    { SUBCCHAINOBJ, VPLIST, PROPERTY },
    { SUBCCHAINOBJ, FINISH, SUGAR },
    { SUBCCHAINOBJ, SUBVPCHAINLOOP, SUGAR },

    { SUBVPCHAINLOOP, SUBCOPCHAIN, SUGAR },
    { SUBCOPCHAIN, SUBCCHAIN, SUGAR },
    { SUBCOPCHAIN, SUBVPCHAIN, REVISE }

  };

// Main method used only for testing
  public static void main(String[] args)
  { Sentence sent = new Sentence("Every animal that lives in a house is a pet.");
    sent.display();
  }

// Constructor
  public Sentence(String arg)
  { textString = arg;
    numberOfCharacters = textString.length();
    makeWords();
    axiom = new OwlTerm(OwlTerm.UNSPECIFIED,2);
    correctSentence = false;
    transduce();
  }

  public String getTextString() {return textString; }
  public OwlTerm getOwlTerm() {return axiom; }
  public boolean sentenceIsCorrect() {return correctSentence; }
  public Word[] getWords() {return word; }
  public int getSpeechAct() {return speechAct; }

  public String getOwlFunctionalSyntax()
  { return axiom.getPrintForm();
  }

// Should be called only if current word is in this sentence.
// If it is not in the sentence, defaults to the first word.
  public Word getCurrentWord(int caretPosition)
  { int wordNumber = 0;
    for (int i=0; i<word.length; i++)
    { if ((word[i].getWordStart() <= caretPosition)
          && (word[i].getWordFinish() >= caretPosition))
      wordNumber = i;
    }
    return word[wordNumber];
  }  

// Return exit state of the last completed word
  public int getCurrentState()
  { int state = START;
    if ((numberOfWords == 1) && (textString.length() > 1))
     { if (textString.charAt(textString.length()-1) == ' ')
       state = word[0].getWordExitState();
     }
    if (numberOfWords > 1)
    { if (word[word.length-1].getWordString().equals("."))
      state = word[word.length-1].getWordExitState();
      else state = word[word.length-2].getWordExitState();
    }
    return state;  
  }

  public static int getRole(int entry, int exit)
  { int r = ERROR;
    for (int i=0; i<role.length; i++)
    if ((role[i][0] == entry) && (role[i][1] == exit)) r = role[i][2];
    return r;
  }

// The Word constructor will assign a syntactic category to each word.
// State and Role variables on each word are assigned by the transducer.
// The vector of user-defined verbs is provided as final argument
  private void makeWords()
  { Tokeniser tokeniser = new Tokeniser(textString);
    numberOfWords = tokeniser.getNumberOfTokens();
    word = new Word[numberOfWords];
    for (int i=0; i<numberOfWords; i++)
    { String istring = tokeniser.getTokenString(i);
      int istart = tokeniser.getTokenStart(i);
      int ifinish = tokeniser.getTokenFinish(i);
      word[i] = new Word(i,istring,istart,ifinish);
    }
  }

// Assign each word in the sentence the state in the network that is
// reached by traversing it (or REJECT if the word violates the grammar).

// At the same time build the OwlTerm interpretation of the sentence in
// axiom (OwlTerm for whole sentence) and register (OwlTerm constituent
// currently under construction).

  private void transduce()
  { int currentState = START;
    int lastState;
    int nArgs;
    OwlTerm register = axiom;
    OwlTerm newTerm, newArg;
    String namepart = "";
    speechAct = ASSERT;
    for (int i=0; i<numberOfWords; i++)
    { lastState = currentState;
      String sss = word[i].getWordString();
      if ((currentState == START) && (speechAct==ASSERT) && speechAct(word[i]))
      { currentState = SPEECHACT;
        if (keyWord(word[i],"Retract")) speechAct = RETRACT;
        else if (keyWord(word[i],"Confirm")) speechAct = CONFIRM;
        else if (keyWord(word[i],"Explain")) speechAct = EXPLAIN;
        else if (keyWord(word[i],"Describe"))
        { speechAct = DESCRIBE;
          currentState = START;
        }
      }
      else if ((currentState == START) && definiteArticle(word[i]))
      { currentState = ISUB;
        register.setFunctor(OwlTerm.CLASS_ASSERTION);
        newTerm = new OwlTerm(OwlTerm.NAMED_INDIVIDUAL,0);
        newTerm.addWord(word[i]);
        register.setArgument(0,newTerm);
      }
      else if ((currentState == START) && properName(word[i]))
      { currentState = SUBJECT;
        register.setFunctor(OwlTerm.CLASS_ASSERTION);
        newTerm = new OwlTerm(OwlTerm.NAMED_INDIVIDUAL,0);
        newTerm.addWord(word[i]);
        register.setArgument(0,newTerm);
      }
      else if ((currentState == START) && pronounWhat(word[i]) && (speechAct==ASSERT))
      { currentState = SUBJECT;
        register.setFunctor(OwlTerm.CLASS_ASSERTION);
        newTerm = new OwlTerm(OwlTerm.NAMED_INDIVIDUAL,0);
        newTerm.addWord(word[i]);
        register.setArgument(0,newTerm);
        speechAct = RETRIEVE;
      }
      else if ((currentState == START) && determinerEvery(word[i]))
      { currentState = CSUB;
        register.setFunctor(OwlTerm.SUB_CLASS_OF);
        newTerm = new OwlTerm(OwlTerm.CLASS,0);
        register.setArgument(0,newTerm);
      }
      else if ((currentState == SPEECHACT) &&
               conjunctionThat(word[i]) &&
               (speechAct == RETRACT))
      { currentState = START;
      }
      else if ((currentState == SPEECHACT) &&
               keyWord(word[i],"whether") &&
               (speechAct == CONFIRM))
      { currentState = START;
      }
      else if ((currentState == SPEECHACT) &&
               keyWord(word[i],"why") &&
               (speechAct == EXPLAIN))
      { currentState = START;
      }
      else if ((currentState == ISUB) && nounPhraseContinuation(word[i]))
      { currentState = SUBJECT;
        newArg = register.getArgument(0);
        newArg.addWord(word[i]);
      }
      else if ((currentState == CSUB) && nounPhraseContinuation(word[i]))
      { currentState = CSUBJECT;
        newArg = register.getArgument(0);
        newArg.addWord(word[i]);
      }
      else if ((currentState == SUBJECT) && nounPhraseContinuation(word[i]))
      { currentState = SUBJECT;
        newArg = register.getArgument(0);
        newArg.addWord(word[i]);
      }
      else if ((currentState == CSUBJECT) && nounPhraseContinuation(word[i]))
      { currentState = CSUBJECT;
        newArg = register.getArgument(0);
        newArg.addWord(word[i]);
      }
      else if ((currentState == SUBJECT) && verbPhraseOpening1(word[i]))
      { currentState = COPULAR;
      }
      else if ((currentState == SUBJECT) && fullStop(word[i]) && (speechAct==DESCRIBE))
      { currentState = FINISH;
        newTerm = unspecifiedClass();
        axiom.setArgument(1,newTerm);
      }
      else if ((currentState == CSUBJECT) && verbPhraseOpening1(word[i]))
      { currentState = COPULAR;
      }
      else if ((currentState == CSUBJECT) && fullStop(word[i]) && (speechAct==DESCRIBE))
      { currentState = FINISH;
        newTerm = unspecifiedClass();
        axiom.setArgument(1,newTerm);
      }
      else if ((currentState == COPULAR) && indefiniteArticle(word[i]))
      { currentState = NPLIST;
        newTerm = new OwlTerm(OwlTerm.CLASS,0);
        register.setArgument(1,newTerm);
      }
      else if ((currentState == NPLIST) && nounPhraseContinuation(word[i]))
      { currentState = COMPLEMENT;
        newArg = register.getArgument(1);
        newArg.addWord(word[i]);
      }
      else if ((currentState == COMPLEMENT) && nounPhraseContinuation(word[i]))
      { currentState = COMPLEMENT;
        newArg = register.getArgument(1);
        newArg.addWord(word[i]);
      }
      else if ((currentState == COMPLEMENT) && fullStop(word[i]))
      { currentState = FINISH;
      }
      else if ((currentState == COPULAR) && verbPhraseContinuation(word[i]))
      { currentState = VPLIST;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i-1]); // incorporate "is" into the property name
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }
      else if ((currentState == SUBJECT) && verbPhraseOpening2(word[i]))
      { currentState = AUX;
      }
      else if ((currentState == CSUBJECT) && verbPhraseOpening2(word[i]))
      { currentState = AUX;
      }
      else if ((currentState == AUX) && verbPhraseContinuation(word[i]))
      { currentState = VPLIST;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i-1]); // incorporate "has" into the property name
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }
      else if ((currentState == SUBJECT) && verbPhraseOpening3(word[i]))
      { currentState = VPLIST;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }
      else if ((currentState == CSUBJECT) && verbPhraseOpening3(word[i]))
      { currentState = VPLIST;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }
      else if ((currentState == CSUBJECT) && conjunctionThat(word[i]))
      { currentState = SUBVPLISTLOOP;
        newTerm = new OwlTerm(OwlTerm.OBJECT_INTERSECTION_OF,2);
        newTerm.setArgument(0,register.getArgument(0));
        register.setArgument(0,newTerm);
        register = newTerm;
      }
      else if ((currentState == VPLIST) && indefiniteArticle(word[i]))
      { currentState = CLIST;
        newTerm = register.getArgument(1);
        newTerm.setFunctor(OwlTerm.OBJECT_SOME_VALUES_FROM);
        newArg = new OwlTerm(OwlTerm.CLASS,0);
        newTerm.setArgument(1,newArg);
      }
      else if ((currentState == VPLIST) && verbPhraseContinuation(word[i]))
      { currentState = VPLIST;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(0);
        newArg.addWord(word[i]);
      }
      else if ((currentState == VPLIST) && definiteArticle(word[i]))
      { currentState = ILIST;
        newTerm = register.getArgument(1);
        newTerm.setFunctor(OwlTerm.OBJECT_HAS_VALUE);
        newArg = new OwlTerm(OwlTerm.NAMED_INDIVIDUAL,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(1,newArg);
      }
      else if ((currentState == ILIST) && nounPhraseContinuation(word[i]))
      { currentState = ILISTOBJ;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(1);
        newArg.addWord(word[i]);
      }
      else if ((currentState == CLIST) && nounPhraseContinuation(word[i]))
      { currentState = CLISTOBJ;
        newTerm = register.getArgument(1);
        if (newTerm.getFunctor()==OwlTerm.CLASS) newTerm.addWord(word[i]);
        else
        { newArg = newTerm.getArgument(1);
          newArg.addWord(word[i]);
        }
      }
      else if ((currentState == VPLIST) && properName(word[i]))
      { currentState = ILISTOBJ;
        newTerm = register.getArgument(1);
        newTerm.setFunctor(OwlTerm.OBJECT_HAS_VALUE);
        newArg = new OwlTerm(OwlTerm.NAMED_INDIVIDUAL,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(1,newArg);
      }
      else if ((currentState == ILISTOBJ) && nounPhraseContinuation(word[i]))
      { currentState = ILISTOBJ;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(1);
        newArg.addWord(word[i]);
      }
      else if ((currentState == CLISTOBJ) && nounPhraseContinuation(word[i]))
      { currentState = CLISTOBJ;
        newTerm = register.getArgument(1);
        if (newTerm.getFunctor()==OwlTerm.CLASS) newTerm.addWord(word[i]);
        else
        { newArg = newTerm.getArgument(1);
          newArg.addWord(word[i]);
        }
      }
      else if ((currentState == ILISTOBJ) && fullStop(word[i]))
      { currentState = FINISH;
      }
      else if ((currentState == CLISTOBJ) && fullStop(word[i]))
      { currentState = FINISH;
      }
      else if ((currentState == ILISTOBJ) && conjunctionAnd(word[i]))
      { currentState = VPLISTLOOP;
        newTerm = new OwlTerm(OwlTerm.OBJECT_INTERSECTION_OF,2);
        newTerm.setArgument(0,register.getArgument(1));
        register.setArgument(1,newTerm);
        register = newTerm;
      }
      else if ((currentState == CLISTOBJ) && conjunctionAnd(word[i]))
      { currentState = VPLISTLOOP;
        newTerm = new OwlTerm(OwlTerm.OBJECT_INTERSECTION_OF,2);
        newTerm.setArgument(0,register.getArgument(1));
        register.setArgument(1,newTerm);
        register = newTerm;
      }



     else if ((currentState == VPLISTLOOP) && verbPhraseOpening1(word[i]))
      { currentState = COPLIST;
      }
     else if ((currentState == VPLISTLOOP) && verbPhraseOpening2(word[i]))
      { currentState = AUXLIST;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }

      else if ((currentState == VPLISTLOOP) && verbPhraseOpening3(word[i]))
      { currentState = VPLIST;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }


      else if ((currentState == COPLIST) && indefiniteArticle(word[i]))
      { currentState = CLIST;
        newTerm = new OwlTerm(OwlTerm.CLASS,0);
        register.setArgument(1,newTerm);
      }
      else if ((currentState == COPLIST) && verbPhraseContinuation(word[i]))
      { currentState = VPLIST;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i-1]); // incorporate "is" into the property name
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }


      else if ((currentState == AUXLIST) && verbPhraseContinuation(word[i]))
      { currentState = VPLIST;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(0);
        newArg.addWord(word[i]);
      }
      else if ((currentState == SUBVPLIST) && indefiniteArticle(word[i]))
      { currentState = SUBCLIST;
        newTerm = register.getArgument(1);
        newTerm.setFunctor(OwlTerm.OBJECT_SOME_VALUES_FROM);
        newArg = new OwlTerm(OwlTerm.CLASS,0);
        newTerm.setArgument(1,newArg);
      }
      else if ((currentState == SUBVPLIST) && verbPhraseContinuation(word[i]))
      { currentState = SUBVPLIST;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(0);
        newArg.addWord(word[i]);
      }
      else if ((currentState == SUBVPLIST) && definiteArticle(word[i]))
      { currentState = SUBILIST;
        newTerm = register.getArgument(1);
        newTerm.setFunctor(OwlTerm.OBJECT_HAS_VALUE);
        newArg = new OwlTerm(OwlTerm.NAMED_INDIVIDUAL,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(1,newArg);
      }
      else if ((currentState == SUBILIST) && nounPhraseContinuation(word[i]))
      { currentState = SUBILISTOBJ;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(1);
        newArg.addWord(word[i]);
      }
      else if ((currentState == SUBCLIST) && nounPhraseContinuation(word[i]))
      { currentState = SUBCLISTOBJ;
        newTerm = register.getArgument(1);
        if (newTerm.getFunctor()==OwlTerm.CLASS) newTerm.addWord(word[i]);
        else
        { newArg = newTerm.getArgument(1);
          newArg.addWord(word[i]);
        }
      }
      else if ((currentState == SUBVPLIST) && properName(word[i]))
      { currentState = SUBILISTOBJ;
        newTerm = register.getArgument(1);
        newTerm.setFunctor(OwlTerm.OBJECT_HAS_VALUE);
        newArg = new OwlTerm(OwlTerm.NAMED_INDIVIDUAL,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(1,newArg);
      }
      else if ((currentState == SUBILISTOBJ) && nounPhraseContinuation(word[i]))
      { currentState = SUBILISTOBJ;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(1);
        newArg.addWord(word[i]);
      }
      else if ((currentState == SUBCLISTOBJ) && nounPhraseContinuation(word[i]))
      { currentState = SUBCLISTOBJ;
        newTerm = register.getArgument(1);
        if (newTerm.getFunctor()==OwlTerm.CLASS) newTerm.addWord(word[i]);
        else
        { newArg = newTerm.getArgument(1);
          newArg.addWord(word[i]);
        }
      }
      else if ((currentState == SUBILISTOBJ) && verbPhraseOpening1(word[i]))
      { currentState = COPULAR;
        register = axiom;
      }
      else if ((currentState == SUBILISTOBJ) && fullStop(word[i]) && (speechAct==DESCRIBE))
      { currentState = FINISH;
        newTerm = unspecifiedClass();
        axiom.setArgument(1,newTerm);
      }
      else if ((currentState == SUBILISTOBJ) && verbPhraseOpening2(word[i]))
      { currentState = AUX;
        register = axiom;
      }
      else if ((currentState == SUBILISTOBJ) && verbPhraseOpening3(word[i]))
      { currentState = VPLIST;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register = axiom;
        register.setArgument(1,newTerm);
      }
      else if ((currentState == SUBCLISTOBJ) && verbPhraseOpening1(word[i]))
      { currentState = COPULAR;
        register = axiom;
      }
      else if ((currentState == SUBCLISTOBJ) && fullStop(word[i]) && (speechAct==DESCRIBE))
      { currentState = FINISH;
        newTerm = unspecifiedClass();
        axiom.setArgument(1,newTerm);
      }
      else if ((currentState == SUBCLISTOBJ) && verbPhraseOpening2(word[i]))
      { currentState = AUX;
        register = axiom;
      }
      else if ((currentState == SUBCLISTOBJ) && verbPhraseOpening3(word[i]))
      { currentState = VPLIST;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register = axiom;
        register.setArgument(1,newTerm);
      }
      else if ((currentState == SUBILISTOBJ) && conjunctionAnd(word[i]))
      { currentState = SUBVPLISTLOOP;
        newTerm = new OwlTerm(OwlTerm.OBJECT_INTERSECTION_OF,2);
        newTerm.setArgument(0,register.getArgument(1));
        register.setArgument(1,newTerm);
        register = newTerm;
      }
      else if ((currentState == SUBCLISTOBJ) && conjunctionAnd(word[i]))
      { currentState = SUBVPLISTLOOP;
        newTerm = new OwlTerm(OwlTerm.OBJECT_INTERSECTION_OF,2);
        newTerm.setArgument(0,register.getArgument(1));
        register.setArgument(1,newTerm);
        register = newTerm;
      }
     else if ((currentState == SUBVPLISTLOOP) && verbPhraseOpening1(word[i]))
      { currentState = SUBCOPLIST;
      }
     else if ((currentState == SUBVPLISTLOOP) && verbPhraseOpening2(word[i]))
      { currentState = SUBAUXLIST;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }
      else if ((currentState == SUBVPLISTLOOP) && verbPhraseOpening3(word[i]))
      { currentState = SUBVPLIST;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }




      else if ((currentState == SUBCOPLIST) && indefiniteArticle(word[i]))
      { currentState = SUBCLIST;
        newTerm = new OwlTerm(OwlTerm.CLASS,0);
        register.setArgument(1,newTerm);
      }
      else if ((currentState == SUBCOPLIST) && verbPhraseContinuation(word[i]))
      { currentState = SUBVPLIST;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i-1]); // incorporate "is" into the property name
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }




      else if ((currentState == SUBAUXLIST) && verbPhraseContinuation(word[i]))
      { currentState = SUBVPLIST;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(0);
        newArg.addWord(word[i]);
      }
      else if ((currentState == COMPLEMENT) && conjunctionThat(word[i]))
      { currentState = VPLISTLOOP;
        newTerm = new OwlTerm(OwlTerm.OBJECT_INTERSECTION_OF,2);
        newTerm.setArgument(0,register.getArgument(1));
        register.setArgument(1,newTerm);
        register = newTerm;
      }
      else if ((currentState == CLISTOBJ) && conjunctionThat(word[i]))
      { currentState = VPCHAINLOOP;
        newTerm = new OwlTerm(OwlTerm.OBJECT_INTERSECTION_OF,2);
        newTerm.setArgument(0,register.getArgument(1).getArgument(1));
        register.getArgument(1).setArgument(1,newTerm);
        register = newTerm;
      }

      else if ((currentState == VPCHAINLOOP) && verbPhraseOpening1(word[i]))
      { currentState = COPCHAIN;
      }

      else if ((currentState == VPCHAINLOOP) && verbPhraseOpening2(word[i]))
      { currentState = AUXCHAIN;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }
      else if ((currentState == VPCHAINLOOP) && verbPhraseOpening3(word[i]))
      { currentState = VPCHAIN;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }


      else if ((currentState == COPCHAIN) && indefiniteArticle(word[i]))
      { currentState = CCHAIN;
        newTerm = new OwlTerm(OwlTerm.CLASS,0);
        register.setArgument(1,newTerm);
      }
      else if ((currentState == COPCHAIN) && verbPhraseContinuation(word[i]))
      { currentState = VPCHAIN;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i-1]); // incorporate "is" into the property name
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }


      else if ((currentState == AUXCHAIN) && verbPhraseContinuation(word[i]))
      { currentState = VPCHAIN;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(0);
        newArg.addWord(word[i]);
      }
      else if ((currentState == VPCHAIN) && verbPhraseContinuation(word[i]))
      { currentState = VPCHAIN;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(0);
        newArg.addWord(word[i]);
      }
      else if ((currentState == VPCHAIN) && definiteArticle(word[i]))
      { currentState = ICHAIN;
        newTerm = register.getArgument(1);
        newTerm.setFunctor(OwlTerm.OBJECT_HAS_VALUE);
        newArg = new OwlTerm(OwlTerm.NAMED_INDIVIDUAL,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(1,newArg);
      }
      else if ((currentState == ICHAIN) && nounPhraseContinuation(word[i]))
      { currentState = ICHAINOBJ;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(1);
        newArg.addWord(word[i]);
      }
      else if ((currentState == VPCHAIN) && indefiniteArticle(word[i]))
      { currentState = CCHAIN;
        newTerm = register.getArgument(1);
        newTerm.setFunctor(OwlTerm.OBJECT_SOME_VALUES_FROM);
        newArg = new OwlTerm(OwlTerm.CLASS,0);
        newTerm.setArgument(1,newArg);
      }
      else if ((currentState == CCHAIN) && nounPhraseContinuation(word[i]))
      { currentState = CCHAINOBJ;
        newTerm = register.getArgument(1);
        if (newTerm.getFunctor()==OwlTerm.CLASS) newTerm.addWord(word[i]);
        else
        { newArg = newTerm.getArgument(1);
          newArg.addWord(word[i]);
        }
      }
      else if ((currentState == VPCHAIN) && properName(word[i]))
      { currentState = ICHAINOBJ;
        newTerm = register.getArgument(1);
        newTerm.setFunctor(OwlTerm.OBJECT_HAS_VALUE);
        newArg = new OwlTerm(OwlTerm.NAMED_INDIVIDUAL,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(1,newArg);
      }
      else if ((currentState == ICHAINOBJ) && nounPhraseContinuation(word[i]))
      { currentState = ICHAINOBJ;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(1);
        newArg.addWord(word[i]);
      }
      else if ((currentState == CCHAINOBJ) && nounPhraseContinuation(word[i]))
      { currentState = CCHAINOBJ;
        newTerm = register.getArgument(1);
        if (newTerm.getFunctor()==OwlTerm.CLASS) newTerm.addWord(word[i]);
        else
        { newArg = newTerm.getArgument(1);
          newArg.addWord(word[i]);
        }
      }
      else if ((currentState == ICHAINOBJ) && fullStop(word[i]))
      { currentState = FINISH;
      }
      else if ((currentState == CCHAINOBJ) && fullStop(word[i]))
      { currentState = FINISH;
      }
      else if ((currentState == CCHAINOBJ) && conjunctionThat(word[i]))
      { currentState = VPCHAINLOOP;
        newTerm = new OwlTerm(OwlTerm.OBJECT_INTERSECTION_OF,2);
        if (register.getArgument(1).getFunctor()==OwlTerm.CLASS)
        { newTerm.setArgument(0,register.getArgument(1));
          register.setArgument(1,newTerm);
        }
        else
        { newTerm.setArgument(0,register.getArgument(1).getArgument(1));
          register.getArgument(1).setArgument(1,newTerm);
        }
        register = newTerm;
      }

      else if ((currentState == SUBCLISTOBJ) && conjunctionThat(word[i]))
      { currentState = SUBVPCHAINLOOP;
        newTerm = new OwlTerm(OwlTerm.OBJECT_INTERSECTION_OF,2);
        if (register.getArgument(1).getFunctor()==OwlTerm.CLASS)
        { newTerm.setArgument(0,register.getArgument(1));
          register.setArgument(1,newTerm);
        }
        else
        { newTerm.setArgument(0,register.getArgument(1).getArgument(1));
          register.getArgument(1).setArgument(1,newTerm);
        }
        register = newTerm;
      }

      else if ((currentState == SUBVPCHAINLOOP) && verbPhraseOpening1(word[i]))
      { currentState = SUBCOPCHAIN;
      }

      else if ((currentState == SUBVPCHAINLOOP) && verbPhraseOpening2(word[i]))
      { currentState = SUBAUXCHAIN;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }
      else if ((currentState == SUBVPCHAINLOOP) && verbPhraseOpening3(word[i]))
      { currentState = SUBVPCHAIN;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }



      else if ((currentState == SUBCOPCHAIN) && indefiniteArticle(word[i]))
      { currentState = SUBCCHAIN;
        newTerm = new OwlTerm(OwlTerm.CLASS,0);
        register.setArgument(1,newTerm);
      }
      else if ((currentState == SUBCOPCHAIN) && verbPhraseContinuation(word[i]))
      { currentState = SUBVPCHAIN;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i-1]); // incorporate "is" into the property name
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register.setArgument(1,newTerm);
      }



      else if ((currentState == SUBAUXCHAIN) && verbPhraseContinuation(word[i]))
      { currentState = SUBVPCHAIN;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(0);
        newArg.addWord(word[i]);
      }
      else if ((currentState == SUBVPCHAIN) && verbPhraseContinuation(word[i]))
      { currentState = SUBVPCHAIN;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(0);
        newArg.addWord(word[i]);
      }
      else if ((currentState == SUBVPCHAIN) && definiteArticle(word[i]))
      { currentState = SUBICHAIN;
        newTerm = register.getArgument(1);
        newTerm.setFunctor(OwlTerm.OBJECT_HAS_VALUE);
        newArg = new OwlTerm(OwlTerm.NAMED_INDIVIDUAL,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(1,newArg);
      }
      else if ((currentState == SUBICHAIN) && nounPhraseContinuation(word[i]))
      { currentState = SUBICHAINOBJ;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(1);
        newArg.addWord(word[i]);
      }
      else if ((currentState == SUBVPCHAIN) && indefiniteArticle(word[i]))
      { currentState = SUBCCHAIN;
        newTerm = register.getArgument(1);
        newTerm.setFunctor(OwlTerm.OBJECT_SOME_VALUES_FROM);
        newArg = new OwlTerm(OwlTerm.CLASS,0);
        newTerm.setArgument(1,newArg);
      }
      else if ((currentState == SUBCCHAIN) && nounPhraseContinuation(word[i]))
      { currentState = SUBCCHAINOBJ;
        newTerm = register.getArgument(1);
        if (newTerm.getFunctor()==OwlTerm.CLASS) newTerm.addWord(word[i]);
        else
        { newArg = newTerm.getArgument(1);
          newArg.addWord(word[i]);
        }
      }
      else if ((currentState == SUBVPCHAIN) && properName(word[i]))
      { currentState = SUBICHAINOBJ;
        newTerm = register.getArgument(1);
        newTerm.setFunctor(OwlTerm.OBJECT_HAS_VALUE);
        newArg = new OwlTerm(OwlTerm.NAMED_INDIVIDUAL,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(1,newArg);
      }
      else if ((currentState == SUBICHAINOBJ) && nounPhraseContinuation(word[i]))
      { currentState = SUBICHAINOBJ;
        newTerm = register.getArgument(1);
        newArg = newTerm.getArgument(1);
        newArg.addWord(word[i]);
      }
      else if ((currentState == SUBCCHAINOBJ) && nounPhraseContinuation(word[i]))
      { currentState = SUBCCHAINOBJ;
        newTerm = register.getArgument(1);
        if (newTerm.getFunctor()==OwlTerm.CLASS) newTerm.addWord(word[i]);
        else
        { newArg = newTerm.getArgument(1);
          newArg.addWord(word[i]);
        }
      }
      else if ((currentState == SUBICHAINOBJ) && verbPhraseOpening1(word[i]))
      { currentState = COPULAR;
        register = axiom;
      }
      else if ((currentState == SUBICHAINOBJ) && fullStop(word[i]) && (speechAct==DESCRIBE))
      { currentState = FINISH;
        newTerm = unspecifiedClass();
        axiom.setArgument(1,newTerm);
      }
      else if ((currentState == SUBICHAINOBJ) && verbPhraseOpening2(word[i]))
      { currentState = AUX;
        register = axiom;
      }
      else if ((currentState == SUBICHAINOBJ) && verbPhraseOpening3(word[i]))
      { currentState = VPLIST;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register = axiom;
        register.setArgument(1,newTerm);
      }
      else if ((currentState == SUBCCHAINOBJ) && verbPhraseOpening1(word[i]))
      { currentState = COPULAR;
        register = axiom;
      }
      else if ((currentState == SUBICHAINOBJ) && fullStop(word[i]) && (speechAct==DESCRIBE))
      { currentState = FINISH;
        newTerm = unspecifiedClass();
        axiom.setArgument(1,newTerm);
      }
      else if ((currentState == SUBCCHAINOBJ) && verbPhraseOpening2(word[i]))
      { currentState = AUX;
        register = axiom;
      }
      else if ((currentState == SUBCCHAINOBJ) && verbPhraseOpening3(word[i]))
      { currentState = VPLIST;
        newTerm = new OwlTerm(OwlTerm.UNSPECIFIED,2);
        newArg = new OwlTerm(OwlTerm.OBJECT_PROPERTY,0);
        newArg.addWord(word[i]);
        newTerm.setArgument(0,newArg);
        register = axiom;
        register.setArgument(1,newTerm);
      }

      else if ((currentState == SUBCCHAINOBJ) && conjunctionThat(word[i]))
      { currentState = SUBVPCHAINLOOP;
        newTerm = new OwlTerm(OwlTerm.OBJECT_INTERSECTION_OF,2);
        if (register.getArgument(1).getFunctor()==OwlTerm.CLASS)
        { newTerm.setArgument(0,register.getArgument(1));
          register.setArgument(1,newTerm);
        }
        else
        { newTerm.setArgument(0,register.getArgument(1).getArgument(1));
          register.getArgument(1).setArgument(1,newTerm);
        }
        register = newTerm;
      }
      else currentState = REJECT;

//      System.out.println(word[i].getWordString() + " " +currentState + " " + register.getPrintForm());
// Assign entry and exit states, role and number to the word just transduced
      word[i].setWordEntryState(lastState);
      word[i].setWordExitState(currentState);
      word[i].setWordRole(getRole(lastState,currentState));

// Reassign roles for current and previous word in five special cases
// First is when current role is same as previous
// Second is when "is" as previous word is revised from SUGAR to PROPERTY
// Third is when "at" as previous word is revised from PROPERTY to SUGAR
// Fourth, when "is defined" previously is revised from SUGAR to PROPERTY
// Fifth, when "The property" previously is revised from INDIVIDUAL to SUGAR

      if (word[i].getWordRole() == RETAIN)
      word[i].setWordRole(word[i-1].getWordRole());
      else if (word[i].getWordRole() == REVISE)
      { word[i].setWordRole(PROPERTY);
        word[i-1].setWordRole(PROPERTY);
      }
      else if (word[i].getWordRole() == SUGARBACK)
      { if (pronounWhat(word[i])) word[i].setWordRole(SUGAR);
        else word[i].setWordRole(INDIVIDUAL);
      }

// Note whether sentence is correctly completed
      if (currentState == FINISH)
      { correctSentence = true;
        if (keyWord(word[i],"?") && (speechAct==ASSERT)) speechAct = CONFIRM;
      }
    }
  }

  private int attributeFunctor(Word word)
  { int functor = 0;
    String s = word.getWordString();
    if (s.equals("functional"))
      functor = OwlTerm.FUNCTIONAL_OBJECT_PROPERTY;
    else if (s.equals("inverse-functional"))
      functor = OwlTerm.INVERSE_FUNCTIONAL_OBJECT_PROPERTY;
    else if (s.equals("symmetric"))
      functor = OwlTerm.SYMMETRIC_OBJECT_PROPERTY;
    else if (s.equals("asymmetric"))
      functor = OwlTerm.ASYMMETRIC_OBJECT_PROPERTY;
    else if (s.equals("reflexive"))
      functor = OwlTerm.REFLEXIVE_OBJECT_PROPERTY;
    else if (s.equals("irreflexive"))
      functor = OwlTerm.IRREFLEXIVE_OBJECT_PROPERTY;
    else if (s.equals("transitive"))
      functor = OwlTerm.TRANSITIVE_OBJECT_PROPERTY;
    return functor;
  }

  private boolean datatype(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.DATATYPE) tval = true;  
    return tval; 
  }

  private OwlTerm unspecifiedClass()
  { OwlTerm oterm = new OwlTerm(OwlTerm.CLASS,0);
    Word word = new Word(0,"what",0,0);
    oterm.addWord(word);
    return oterm;
  }

  private boolean propertyAttribute(Word word)
  { boolean tval = false;
    String s = word.getWordString();
    if (s.equals("functional") ||
        s.equals("inverse-functional") ||
        s.equals("symmetric") ||
        s.equals("asymmetric") ||
        s.equals("reflexive") ||
        s.equals("irreflexive") ||
        s.equals("transitive") ||
        s.equals("[attribute]"))
    tval = true;
    return tval;
  }

  private int relationFunctor(Word word)
  { int functor = 0;
    String s = word.getWordString();
    if (s.equals("disjoint"))
      functor = OwlTerm.DISJOINT_OBJECT_PROPERTIES;
    else if (s.equals("equivalent"))
      functor = OwlTerm.EQUIVALENT_OBJECT_PROPERTIES;
    else if (s.equals("inverse") || s.equals("inverses"))
      functor = OwlTerm.INVERSE_OBJECT_PROPERTIES;
    return functor;
  }

  private boolean propertyRelation(Word word)
  { boolean tval = false;
    String s = word.getWordString();
    if (s.equals("disjoint") ||
        s.equals("equivalent") ||
        s.equals("inverse") ||
        s.equals("inverses"))
    tval = true;
    return tval;
  }

  private int dataRelationFunctor(Word word)
  { int functor = 0;
    String s = word.getWordString();
    if (s.equals("disjoint"))
      functor = OwlTerm.DISJOINT_DATA_PROPERTIES;
    else if (s.equals("equivalent"))
      functor = OwlTerm.EQUIVALENT_DATA_PROPERTIES;
    return functor;
  }

  private boolean dataPropertyRelation(Word word)
  { boolean tval = false;
    String s = word.getWordString();
    if (s.equals("disjoint") || s.equals("equivalent"))
    tval = true;
    return tval;
  }

  private boolean singularInteger(Word word)
  { boolean tval = false;
    String s = word.getWordString();
    if (s.equals("one") || s.equals("1")) tval = true;
    return tval;
  }

  private boolean keyWord(Word word, String key)
  { boolean tval = false;
    if (word.getWordString().equals(key)) tval = true;  
    return tval; 
  }

  private boolean pronoun(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.PRONOUN) tval = true;  
    return tval; 
  }

  private boolean pronounWhat(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.PRONOUN_WHAT) tval = true;  
    return tval; 
  }

  private boolean speechAct(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.SPEECH_ACT) tval = true;  
    return tval; 
  }

  private boolean definiteArticle(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.DEFINITE_ARTICLE) tval = true;  
    return tval; 
  }

  private boolean indefiniteArticle(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.INDEFINITE_ARTICLE) tval = true;  
    return tval; 
  }

  private boolean determinerEvery(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.INDEFINITE_ARTICLE) tval = true;
    if (word.getWordCategory() == Word.DETERMINER_EVERY) tval = true; 
    return tval; 
  }

  private boolean determinerNo(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.DETERMINER_NO) tval = true; 
    return tval; 
  }

  private boolean determinerAny(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.DETERMINER_ANY) tval = true; 
    return tval; 
  }

  private boolean quantifierAnything(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.QUANTIFIER_ANYTHING) tval = true; 
    return tval; 
  }

  private boolean quantifierSomething(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.QUANTIFIER_SOMETHING) tval = true; 
    return tval; 
  }

  private boolean quantifierSome(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.QUANTIFIER_SOME) tval = true; 
    return tval; 
  }

  private boolean quantifierOnly(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.QUANTIFIER_ONLY) tval = true; 
    return tval; 
  }

  private boolean quantifierMost(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.QUANTIFIER_MOST) tval = true; 
    return tval; 
  }

  private boolean quantifierLeast(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.QUANTIFIER_LEAST) tval = true; 
    return tval; 
  }

  private boolean quantifierExactly(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.QUANTIFIER_EXACTLY) tval = true; 
    return tval; 
  }

  private boolean properName(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.PROPER_NAME) tval = true;  
    return tval;
  }

  private boolean literal(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.STRING) tval = true;  
    if (word.getWordCategory() == Word.INTEGER) tval = true;  
    if (word.getWordCategory() == Word.DECIMAL) tval = true;  
    return tval;
  }

  private boolean integer(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.INTEGER) tval = true;  
    return tval;
  }

  private boolean conjunctionAnd(Word word)
  { boolean tval = false;
    if (word.getWordString().equals("and")) tval = true;  
    return tval; 
  }

  private boolean conjunctionOr(Word word)
  { boolean tval = false;
    if (word.getWordString().equals("or")) tval = true;  
    return tval; 
  }

  private boolean conjunctionThat(Word word)
  { boolean tval = false;
    if (word.getWordString().equals("that")) tval = true;  
    return tval; 
  }

  private boolean conjunctionNot(Word word)
  { boolean tval = false;
    if (word.getWordString().equals("not")) tval = true;  
    return tval; 
  }

  private boolean openAngleBracket(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.OPEN_ANGLE_BRACKET) tval = true;  
    return tval; 
  }

  private boolean closeAngleBracket(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.CLOSE_ANGLE_BRACKET) tval = true;  
    return tval; 
  }

  private boolean nounPhraseContinuation(Word word)
  { boolean tval = false;
    if (definiteArticle(word) == true) tval = true;
    if (word.getWordCategory() == Word.NOUN) tval = true;  
    if (word.getWordCategory() == Word.PROPER_NAME) tval = true;  
    if (word.getWordCategory() == Word.STRING) tval = true;  
    if (word.getWordCategory() == Word.INTEGER) tval = true;  
    if (word.getWordCategory() == Word.DECIMAL) tval = true;  
    if (word.getWordCategory() == Word.PREPOSITION) tval = true;
    return tval;
  }

 private boolean verbPhraseOpening1(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.AUXILIARY_IS) tval = true;
    return tval; 
  }

 private boolean verbPhraseOpening2(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.AUXILIARY_HAS) tval = true;
    return tval; 
  }

 private boolean verbPhraseOpening3(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.VERB) tval = true;
    return tval; 
  }

  private boolean verbPhraseOpening12(Word word)
  { boolean tval = false;
    if (verbPhraseOpening1(word) || verbPhraseOpening2(word)) tval = true;
    return tval; 
  }

  private boolean verbPhraseContinuation(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.NOUN) tval = true;  
    if (word.getWordCategory() == Word.PREPOSITION) tval = true;
    return tval;
  }

  private boolean fullStop(Word word)
  { boolean tval = false;
    if (word.getWordCategory() == Word.FULL_STOP) tval = true;  
    return tval;
  }

// On encountering a datatype rather than a class name, rename the functor
// so that it says DATA in place of OBJECT
  private void objectToDataSwitch(OwlTerm term)
  { int functor = term.getFunctor();
    if (functor == OwlTerm.OBJECT_SOME_VALUES_FROM) term.setFunctor(OwlTerm.DATA_SOME_VALUES_FROM);
    else if (functor == OwlTerm.OBJECT_ALL_VALUES_FROM) term.setFunctor(OwlTerm.DATA_ALL_VALUES_FROM);
    else if (functor == OwlTerm.OBJECT_EXACT_CARDINALITY) term.setFunctor(OwlTerm.DATA_EXACT_CARDINALITY);
    else if (functor == OwlTerm.OBJECT_MAX_CARDINALITY) term.setFunctor(OwlTerm.DATA_MAX_CARDINALITY);
    else if (functor == OwlTerm.OBJECT_MIN_CARDINALITY) term.setFunctor(OwlTerm.DATA_MIN_CARDINALITY);
    else if (functor == OwlTerm.OBJECT_PROPERTY_RANGE) term.setFunctor(OwlTerm.DATA_PROPERTY_RANGE);
    else if (functor == OwlTerm.OBJECT_PROPERTY) term.setFunctor(OwlTerm.DATA_PROPERTY);
  }

  public void display()
  { System.out.println("Text: " + textString);
    System.out.println("Number of characters: " + numberOfCharacters);
    for (int i=0; i<word.length; i++) word[i].display();
    System.out.println("Number of words: " + numberOfWords);
    System.out.println(speechAct);
    if (speechAct==RETRACT) System.out.println("Speech act: Retract");
    else if (speechAct==CONFIRM) System.out.println("Speech act: Confirm");
    else if (speechAct==EXPLAIN) System.out.println("Speech act: Explain");
    else if (speechAct==RETRIEVE) System.out.println("Speech act: Retrieve");
    else if (speechAct==DESCRIBE) System.out.println("Speech act: Describe");
    else System.out.println("Speech act: Assert");
    axiom.display();
  }
}
