
/*
Performs tokenisation on a line of text.

Written by Richard Power
Copyright (c) 2013 by Open University

This package is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Library General Public License for more details.

=================================================================

When an instance is constructed by giving it a string representing a
line of text, it makes available a tokenisation of the string through
arrays containing the tokens (strings), and their start and end positions
in the original string.

If a line of text is empty, it is treated as a single word under
construction, currently empty, rather than as no words at all. This
means that we need to measure starting and finishing position so that
Start + Length = Finish. The finish is therefore the caret position
following the final character of the word. If the word "cat" starts
at character 5 in a string, it occupies characters 5-7 and its finish
is at character 8. The empty word "" beginning at 5 also finishes at 5.

Some test cases are as follows:

Test case 0
Input text line = ""
Line length = 0
numberOfTokens = 1
Token 0 has string="", start=0, finish=0

Test case 1
Input text line = " "
Line length = 1
numberOfTokens = 1
Token 0 has string="", start=0, finish=0

Test case 2
Input text line = "abc"
Line length = 3
numberOfTokens = 1
Token 0 has string="abc", start=0, finish=3

Test case 3
Input text line = " abc"
Line length = 4
numberOfTokens = 1
Token 0 has string="abc", start=1, finish=4

Test case 4
Input text line = "abc "
Line length = 4
numberOfTokens = 2
Token 0 has string="abc", start=0, finish=3
Token 1 has string="", start=4, finish=4

Test case 5
Input text line = "abc  "
Line length = 5
numberOfTokens = 2
Token 0 has string="abc", start=0, finish=3
Token 1 has string="", start=4, finish=4

Test case 6
Input text line = "abc."
Line length = 4
numberOfTokens = 2
Token 0 has string="abc", start=0, finish=3
Token 1 has string=".", start=3, finish=4

Test case 7
Input text line = "abc  def.. "
Line length = 11
numberOfTokens = 3
Token 0 has string="abc", start=0, finish=3
Token 1 has string="def", start=5, finish=8
Token 2 has string=".", start=8, finish=9

Test case 8
Input text line = "abc<def> <g hi>."
Length: 16
Number of tokens: 9
Token 0: 0-3 abc
Token 1: 3-4 <
Token 2: 4-7 def
Token 3: 7-8 >
Token 4: 9-10 <
Token 5: 10-11 g
Token 6: 12-14 hi
Token 7: 14-15 >
Token 8: 15-16 .

*/

import java.util.Vector;
import java.util.Collections;
import java.lang.String;

public class Tokeniser
{ public String lineString;
  public int lineLength;
  public Vector<Token> token;

// Network states
  public static final int REJECT = 0;
  public static final int START = 1;
  public static final int WORD = 2;
  public static final int STRING = 3;
  public static final int NUMBER = 4;
  public static final int DECIMAL = 5;
  public static final int POINT = 6;
  public static final int FINISH = 7;
// Further states for comments and metadata
  public static final int STARTCOMMENT = 8;
  public static final int STARTMETADATA = 9;
  public static final int STARTMETAVALUE = 10;
  public static final int COMMENT = 11;
  public static final int METADATA = 12;
  public static final int METAVALUE = 13;

// Input categories classifiying the next character
  public static final int END_OF_STRING = 0;
  public static final int SPACE = 1;
  public static final int FULL_STOP = 2;
  public static final int DOUBLE_QUOTE = 3;
  public static final int DIGIT = 4;
  public static final int LETTER = 5;
  public static final int PERCENTAGE = 6;
  public static final int HASH = 7;
  public static final int OPEN_ANGLE_BRACKET = 8;
  public static final int CLOSE_ANGLE_BRACKET = 9;
  public static final int INCORRECT_CHARACTER = 10;
  public static final int QUESTION_MARK = 11;

// Test cases
  private static String test[] =
    { "",
      " ",
      "abc",
      " abc",
      "abc ",
      "abc  ",
      "abc.",
      "abc  def.. ",
      "abc<def> <g hi>.",
      "abc 123 45.67 5.",
      "abc123 3ab\"hallo\"5.3a.",
      "A dog has age 3.",
      "abc \"def 123 0.5 ghi\" xyz",
      "% Comment string",
      "# VERB take takes",
      "# OntologyIRI http://ontology.org/ont34"
    };

  public static void main(String[] args) {
    System.out.println("----------");
    for (int i=0; i<test.length; i++)
    { Tokeniser tokens = new Tokeniser(test[i]);
      System.out.println("Test case " + i);
      tokens.display();
      System.out.println("----------");
    }
  }

  public Tokeniser(String stringArg)
  { lineString = stringArg;
    lineLength = lineString.length();
    tokenise();
  }

  public int getNumberOfTokens() {return token.size(); }
  public String getTokenString(int i) {return token.elementAt(i).getTokenString(); }
  public int getTokenStart(int i) {return token.elementAt(i).getTokenStart(); }
  public int getTokenFinish(int i) {return token.elementAt(i).getTokenFinish(); }

// Inner class for tokens

  public class Token
  { private String tokenString;
    private int tokenStart, tokenFinish;

    public Token(String arg1, int arg2, int arg3)
    { tokenString = arg1;
      tokenStart = arg2;
      tokenFinish = arg3;
    }

    public void setTokenString(String s) { tokenString = s; }
    public void setTokenStart(int i) { tokenStart = i; }
    public void setTokenFinish(int i) { tokenFinish = i; }

    public String getTokenString() { return tokenString; }
    public int getTokenStart() { return tokenStart; }
    public int getTokenFinish() { return tokenFinish; }
  }

// Transduce the string

  private void tokenise()
  { token = new Vector<Token>();
    int index = 0;
    int currentState = START;
    int ntokens = 0;
    int input;
    Token newToken;
    while ((currentState != FINISH) && (currentState != REJECT))
    { input = inputCategory(index);
 // Transitions from START
      if ((currentState == START) && (input == END_OF_STRING))
      { currentState = FINISH;
        if ((ntokens == 0) || (!token.elementAt(ntokens-1).getTokenString().equals(".")))
        { newToken = new Token("",index,index);
          ntokens = ntokens + 1;
          token.addElement(newToken);
        }
      }
      else if ((currentState == START) && (input == SPACE))
      { currentState = START;
      }
      else if ((currentState == START) && (input == FULL_STOP))
      { currentState = START;
        newToken = new Token(".",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == START) && (input == OPEN_ANGLE_BRACKET))
      { currentState = START;
        newToken = new Token("<",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == START) && (input == CLOSE_ANGLE_BRACKET))
      { currentState = START;
        newToken = new Token(">",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == START) && (input == PERCENTAGE))
      { currentState = STARTCOMMENT;
        newToken = new Token("%",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == START) && (input == HASH))
      { currentState = STARTMETADATA;
        newToken = new Token("#",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == START) && (input == DOUBLE_QUOTE))
      { currentState = STRING;
        newToken = new Token("\"",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == START) && (input == DIGIT))
      { currentState = NUMBER;
        newToken = new Token(Character.toString(lineString.charAt(index)),index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == START) && (input == LETTER))
      { currentState = WORD;
        newToken = new Token(Character.toString(lineString.charAt(index)),index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
// Transitions from WORD
      else if ((currentState == WORD) && (input == END_OF_STRING))
      { currentState = FINISH;
      }
      else if ((currentState == WORD) && (input == SPACE))
      { currentState = START;
      }
      else if ((currentState == WORD) && (input == FULL_STOP))
      { currentState = START;
        newToken = new Token(".",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == WORD) && (input == QUESTION_MARK))
      { currentState = START;
        newToken = new Token("?",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == WORD) && (input == OPEN_ANGLE_BRACKET))
      { currentState = START;
        newToken = new Token("<",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == WORD) && (input == CLOSE_ANGLE_BRACKET))
      { currentState = START;
        newToken = new Token(">",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == WORD) && (input == DOUBLE_QUOTE))
      { currentState = STRING;
        newToken = new Token("\"",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == WORD) && (input == DIGIT))
      { currentState = WORD;
        updateToken(ntokens-1,index);
      }
      else if ((currentState == WORD) && (input == LETTER))
      { currentState = WORD;
        updateToken(ntokens-1,index);
      }
// Transitions from STRING
      else if ((currentState == STRING) && (input == END_OF_STRING))
      { currentState = FINISH;
      }
      else if ((currentState == STRING) && (input == SPACE))
      { currentState = STRING;
        updateToken(ntokens-1,index);
      }
      else if ((currentState == STRING) && (input == FULL_STOP))
      { currentState = STRING;
        updateToken(ntokens-1,index);
      }
      else if ((currentState == STRING) && (input == OPEN_ANGLE_BRACKET))
      { currentState = STRING;
        updateToken(ntokens-1,index);
      }
      else if ((currentState == STRING) && (input == CLOSE_ANGLE_BRACKET))
      { currentState = STRING;
        updateToken(ntokens-1,index);
      }
      else if ((currentState == STRING) && (input == DOUBLE_QUOTE))
      { currentState = START;
        updateToken(ntokens-1,index);
      }
      else if ((currentState == STRING) && (input == DIGIT))
      { currentState = STRING;
        updateToken(ntokens-1,index);
      }
      else if ((currentState == STRING) && (input == LETTER))
      { currentState = STRING;
        updateToken(ntokens-1,index);
      }
// Transitions from NUMBER
      else if ((currentState == NUMBER) && (input == END_OF_STRING))
      { currentState = FINISH;
      }
      else if ((currentState == NUMBER) && (input == SPACE))
      { currentState = START;
      }
      else if ((currentState == NUMBER) && (input == FULL_STOP))
      { currentState = POINT;
      }
      else if ((currentState == NUMBER) && (input == OPEN_ANGLE_BRACKET))
      { currentState = START;
        newToken = new Token("<",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == NUMBER) && (input == CLOSE_ANGLE_BRACKET))
      { currentState = START;
        newToken = new Token(">",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == NUMBER) && (input == DOUBLE_QUOTE))
      { currentState = STRING;
        newToken = new Token("\"",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == NUMBER) && (input == DIGIT))
      { currentState = NUMBER;
        updateToken(ntokens-1,index);
      }
      else if ((currentState == NUMBER) && (input == LETTER))
      { currentState = WORD;
        updateToken(ntokens-1,index);
      }
// Transitions from POINT
      else if ((currentState == POINT) && (input == END_OF_STRING))
      { currentState = FINISH;
        newToken = new Token(".",index-1,index);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == POINT) && (input == SPACE))
      { currentState = START;
        newToken = new Token(".",index-1,index);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == POINT) && (input == DOUBLE_QUOTE))
      { currentState = STRING;
        newToken = new Token(".",index-1,index);
        ntokens = ntokens + 1;
        token.addElement(newToken);
        newToken = new Token("\"",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == POINT) && (input == FULL_STOP))
      { currentState = START;
        newToken = new Token(".",index-1,index);
        ntokens = ntokens + 1;
        token.addElement(newToken);
        newToken = new Token(".",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == POINT) && (input == OPEN_ANGLE_BRACKET))
      { currentState = START;
        newToken = new Token("<",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == POINT) && (input == CLOSE_ANGLE_BRACKET))
      { currentState = START;
        newToken = new Token(">",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == POINT) && (input == DIGIT))
      { currentState = DECIMAL;
        updateToken(ntokens-1,index-1);
        updateToken(ntokens-1,index);
      }
      else if ((currentState == POINT) && (input == LETTER))
      { currentState = WORD;
        newToken = new Token(".",index-1,index);
        ntokens = ntokens + 1;
        token.addElement(newToken);
        newToken = new Token(Character.toString(lineString.charAt(index)),index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
// Transitions from DECIMAL
      else if ((currentState == DECIMAL) && (input == END_OF_STRING))
      { currentState = FINISH;
      }
      else if ((currentState == DECIMAL) && (input == SPACE))
      { currentState = START;
      }
      else if ((currentState == DECIMAL) && (input == FULL_STOP))
      { currentState = START;
        newToken = new Token(".",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == DECIMAL) && (input == OPEN_ANGLE_BRACKET))
      { currentState = START;
        newToken = new Token("<",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == DECIMAL) && (input == CLOSE_ANGLE_BRACKET))
      { currentState = START;
        newToken = new Token(">",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == DECIMAL) && (input == DOUBLE_QUOTE))
      { currentState = STRING;
        newToken = new Token("\"",index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
      else if ((currentState == DECIMAL) && (input == DIGIT))
      { currentState = DECIMAL;
        updateToken(ntokens-1,index);
      }
      else if ((currentState == DECIMAL) && (input == LETTER))
      { currentState = WORD;
        newToken = new Token(Character.toString(lineString.charAt(index)),index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
// Transitions from STARTCOMMENT
      else if ((currentState == STARTCOMMENT) && (input == SPACE))
      { currentState = STARTCOMMENT;
      }
      else if ((currentState == STARTCOMMENT) && (input != END_OF_STRING))
      { currentState = COMMENT;
        newToken = new Token(Character.toString(lineString.charAt(index)),index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
// Transitions from COMMENT
      else if ((currentState == COMMENT) && (input == END_OF_STRING))
      { currentState = FINISH;
      }
      else if (currentState == COMMENT)
      { currentState = COMMENT;
        updateToken(ntokens-1,index);
      }
// Transitions from STARTMETADATA
      else if ((currentState == STARTMETADATA) && (input == SPACE))
      { currentState = STARTMETADATA;
      }
      else if ((currentState == STARTMETADATA) && (input == LETTER))
      { currentState = METADATA;
        newToken = new Token(Character.toString(lineString.charAt(index)),index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
// Transitions from METADATA
      else if ((currentState == METADATA) && (input == SPACE))
      { currentState = STARTMETAVALUE;
      }
      else if ((currentState == METADATA) && (input == LETTER))
      { currentState = METADATA;
        updateToken(ntokens-1,index);
      }
// Transitions from STARTMETAVALUE
      else if ((currentState == STARTMETAVALUE) && (input == SPACE))
      { currentState = STARTMETAVALUE;
      }
      else if ((currentState == STARTMETAVALUE) && (input == LETTER))
      { currentState = METAVALUE;
        newToken = new Token(Character.toString(lineString.charAt(index)),index,index+1);
        ntokens = ntokens + 1;
        token.addElement(newToken);
      }
// Transitions from METAVALUE
      else if ((currentState == METAVALUE) && (input == END_OF_STRING))
      { currentState = FINISH;
      }
      else if (currentState == METAVALUE)
      { currentState = METAVALUE;
        updateToken(ntokens-1,index);
      }

      else currentState = REJECT;

      index++;
    }
  }

// Adds the character at the indexed position in the line to the string for
// the indexed token, and updates the finish position of this token.
  private void updateToken(int tokenIndex, int lineIndex)
  { Token tok = token.elementAt(tokenIndex);
    String s = tok.getTokenString(); 
    s = s + Character.toString(lineString.charAt(lineIndex));
    tok.setTokenString(s);
    tok.setTokenFinish(lineIndex+1);
  }

  private int inputCategory(int index)
  { int category;
    if (index >= lineLength) category = END_OF_STRING;
    else category = charCategory(lineString.charAt(index));
    return category;
  }

  private int charCategory(char c)
  { int category;
    if (c == ' ') category = SPACE;
    else if (c == '.') category = FULL_STOP;
    else if (c == '?') category = QUESTION_MARK;
    else if (c == '"') category = DOUBLE_QUOTE;
    else if (c == '%') category = PERCENTAGE;
    else if (c == '#') category = HASH;
    else if (c == '<') category = OPEN_ANGLE_BRACKET;
    else if (c == '>') category = CLOSE_ANGLE_BRACKET;
    else if (Character.isDigit(c)) category = DIGIT;
    else category = LETTER;
    return category;
  }

  public void display()
  { int ntokens = getNumberOfTokens();
    System.out.println("String: " + lineString);
    System.out.println("Length: " + lineLength);
    System.out.println("Number of tokens: " + ntokens);
    for (int i=0; i<ntokens; i++)
    System.out.println("Token " + i + ": " + getTokenStart(i)
                       + "-" + getTokenFinish(i) 
                       + " " + getTokenString(i));
  }
    
}
